"""Generalized standard material

Formulation of a generic framework for the derivation of constitutive models capturing 
dissipative processes in response of inelastic materials to mechanical and environmental 
loading. This framework aspires to combine the high-level mathematical formulation of the
thermodynamic potentials with an efficient execution of the automatically derived evolution
equations in the computational environment.
"""

from typing import TYPE_CHECKING, Any, Tuple, List, Union, cast, Dict, Optional, Callable
import traits.api as tr
import sympy as sp
import numpy as np
import numpy.typing as npt
import dill
import os
import functools
from functools import cached_property  # Python 3.8+

# Use TYPE_CHECKING to avoid runtime overhead and typing conflicts with SymPy
if TYPE_CHECKING:
    # Tell type checkers what the properties should return
    _sig_lambdified: Callable[..., npt.NDArray[np.float64]]
    phi_: Any  # SymPy expression
    Y_: Any    # SymPy matrix
    Phi_: Any  # SymPy matrix

    from typing import Any as SymExpr  # Simplified for type checking
    SymMatrix = Any
    SymDict = Dict[Any, Any]
else:
    # At runtime, use actual SymPy types without constraints
    SymExpr = Union[sp.Expr, sp.Basic, sp.Symbol]
    SymMatrix = sp.Matrix
    SymDict = Dict[SymExpr, SymExpr]

def get_dirac_delta(x: Any, x_0: float = 0) -> int:
    return 0

numpy_dirac = [{'DiracDelta': get_dirac_delta}, 'numpy']

# Directory to store the serialized symbolic instances
CACHE_DIR = '_lambdified_cache'

class GSMEngine(tr.HasTraits):
    """Generalized Standard Material

    The class definition consists of symbolic expressions and numerical execution methods
    for thermodynamically consistent material modeling.
    """

    name = tr.Str('unnamed')
    eps_vars = tr.Any()
    sig_vars = tr.Any()
    T_var = tr.Any()
    Eps_vars = tr.Tuple()
    Sig_vars = tr.Tuple()
    m_params = tr.Tuple()
    m_param_subs = tr.Dict({})
    F_expr = tr.Any()
    pi_expr = tr.Any(default_value=0)
    Sig_signs = tr.Array()
    sig_sign = tr.Float(1)
    gamma_mech_sign = tr.Int(-1)
    dot_Eps_bounds_expr = tr.Any(sp.S.Zero)
    phi_ext_expr = tr.Any(sp.S.Zero)
    f_expr = tr.Any(sp.S.Zero)
    d_t = sp.Symbol(r'\mathrm{d}t', real=True)
    h_k = tr.List(tr.Any, value=[])
    vp_on = tr.Bool(True)
    update_at_k = tr.Bool(True)
    k_max = tr.Int(10)

    # Properties - let SymPy handle its own typing internally
    eps_a = tr.Property()
    @tr.cached_property
    def _get_eps_a(self):
        return sp.BlockMatrix([
            var if isinstance(var, sp.Matrix) and var.shape[1] == 1 else sp.Matrix([[var]]) 
            for var in self.eps_vars
        ]).as_explicit()

    dot_eps_a = tr.Property()
    @tr.cached_property
    def _get_dot_eps_a(self):
        return sp.Matrix([sp.Symbol(f'\\dot{{{var.name}}}') for var in self.eps_a])

    sig_a = tr.Property()
    @tr.cached_property
    def _get_sig_a(self):
        return sp.BlockMatrix([
            var if isinstance(var, sp.Matrix) and var.shape[1] == 1 else sp.Matrix([[var]]) 
            for var in self.sig_vars
        ]).as_explicit()

    dot_sig_a = tr.Property()
    @tr.cached_property
    def _get_dot_sig_a(self):
        return sp.Matrix([sp.Symbol(f'\\dot{{{var.name}}}') for var in self.sig_vars])

    # For complex SymPy operations, don't over-constrain with typing
    lam_phi_f_ = tr.Property()
    @tr.cached_property
    def _get_lam_phi_f_(self):
        if self.phi_ == sp.S.Zero:
            return ([], sp.S.Zero, sp.S.Zero)
        lam_phi = sp.Symbol(r'\lambda_{\mathrm{\phi}}', real=True)
        return ([lam_phi], lam_phi * (-self.phi_), self.f_expr)

    n_lam = tr.Property()
    @tr.cached_property
    def _get_n_lam(self):
        lam_phi_k, _, _ = self.lam_phi_f_
        return len(lam_phi_k)

    H_Lam = tr.Property()
    @tr.cached_property
    def _get_H_Lam(self):
        dot_Lam = [sp.Symbol(f'\\dot{{\\lambda}}_{{{k}}}', real=True) for k in range(len(self.h_k))]
        delta_Lam = sp.Matrix([sp.Symbol(f'\\Delta\\lambda_{{{k}}}', real=True) for k in range(len(self.h_k))])
        dot_lam_sum = sum(l * h for l, h in zip(dot_Lam, self.h_k)) if self.h_k else sp.S.Zero
        return (dot_Lam, delta_Lam, dot_lam_sum)

    n_Lam = tr.Property()
    @tr.cached_property
    def _get_n_Lam(self):
        return len(self.h_k)

    # Internal variable representations - let SymPy handle its types
    Eps_list = tr.Property()
    @tr.cached_property
    def _get_Eps_list(self):
        return [Eps_i.T for Eps_i in self.Eps_vars]

    Eps = tr.Property()
    @tr.cached_property
    def _get_Eps(self):
        return sp.BlockMatrix(self.Eps_list).T
    
    n_Eps = tr.Property()
    @tr.cached_property
    def _get_n_Eps(self):
        return len(self.Eps.as_explicit())

    _Eps_as_array_lambdified = tr.Property()
    @tr.cached_property
    def _get__Eps_as_array_lambdified(self):
        return sp.lambdify(self.Eps.blocks, self.Eps.as_explicit())

    def Eps_as_array(self, arr: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]:
        return self._Eps_as_array_lambdified(arr)[:, 0]

    _Eps_as_blocks_lambdified = tr.Property()
    @tr.cached_property
    def _get__Eps_as_blocks_lambdified(self):
        return sp.lambdify(self.Eps.as_explicit(), self.Eps_list)

    def Eps_as_blocks(self, arr: npt.NDArray[np.float64]) -> List[Any]:
        return [Eps_i[0] for Eps_i in self._Eps_as_blocks_lambdified(arr)]
       
    dot_Eps_list = tr.Property()
    @tr.cached_property
    def _get_dot_Eps_list(self):
        return [
            sp.Matrix(eps_var.shape[0], eps_var.shape[1], 
                    lambda i, j: sp.Symbol(name=f'\\dot{{{eps_var[i, j].name}}}'))
            for eps_var in self.Eps_list
        ]
    
    dot_Eps = tr.Property()
    @tr.cached_property
    def _get_dot_Eps(self):
        return sp.BlockMatrix(self.dot_Eps_list).T

    # Conjugate state variable representations
    Sig_list = tr.Property()
    @tr.cached_property
    def _get_Sig_list(self):
        return [Sig_i.T for Sig_i in self.Sig_vars]

    Sig = tr.Property()
    @tr.cached_property
    def _get_Sig(self):
        return sp.BlockMatrix([Sig_i.T for Sig_i in self.Sig_vars]).T

    dot_Sig_list = tr.Property()
    @tr.cached_property
    def _get_dot_Sig_list(self):
        return [
            sp.Matrix(sig_var.shape[0], sig_var.shape[1], 
                    lambda i, j: sp.Symbol(name=f'\\dot{{{sig_var[i, j].name}}}'))
            for sig_var in self.Sig_list
        ]
    
    dot_Sig = tr.Property()
    @tr.cached_property
    def _get_dot_Sig(self):
        return sp.BlockMatrix(self.dot_Sig_list).T

    dF_dEps_ = tr.Property()
    @tr.cached_property
    def _get_dF_dEps_(self):
        return sp.BlockMatrix([sp.simplify(sp.diff(self.F_expr, var)).T for var in self.Eps.blocks]).T.subs(self.m_param_subs)

    # Numerical methods - type the interfaces clearly
    def get_sig_a(self, eps: Union[float, npt.NDArray[np.float64]], Eps: Optional[npt.NDArray[np.float64]] = None, *args: float) -> Union[float, npt.NDArray[np.float64]]:
        """Calculate the displacement for a given stress level"""
        eps_sp_ = np.moveaxis(np.atleast_1d(eps), -1, 0)
        if Eps is not None:
            Eps_sp_ = np.moveaxis(Eps, -1, 0)
        else:
            Eps_sp_ = np.zeros((1, self.n_Eps))
        sig_sp_ = self._sig_a_lambdified(eps_sp_, Eps_sp_, *args)  # type: ignore[misc]
        sig_sp_ = sig_sp_.reshape(eps_sp_.shape)
        return np.moveaxis(sig_sp_, 0, -1)

    # For new properties, use Python's cached_property instead of traits
    @cached_property  
    def _sig_a_lambdified(self) -> Callable[..., npt.NDArray[np.float64]]:
        return sp.lambdify((self.eps_vars[0], self.Eps.as_explicit()) + self.m_params + ('*args',), 
                           self.sig_a_, numpy_dirac, cse=True)

    # Keep existing traits properties as-is, just ignore typing for them
    phi_ = tr.Property()  # type: ignore[misc]
    @tr.cached_property
    def _get_phi_(self):
        return (self.f_expr + self.phi_ext_expr).subs(self.m_param_subs)

    sig_a_ = tr.Property()
    @tr.cached_property
    def _get_sig_a_(self):
        return self.F_expr.diff(self.eps_a)

    subs_sig_eps = tr.Property()
    @tr.cached_property
    def _get_subs_sig_eps(self):
        return dict(zip(self.sig_a, self.sig_a_))  # type: ignore[arg-type]

    # More numerical interface methods...
    def get_dDiss_dEps(self, eps: Any, T: Any, Eps: Any, Sig: Any, *args: Any) -> npt.NDArray[np.float64]:
        """Calculate the derivative of the dissipation rate with respect to internal variables"""
        return self._dDiss_dEps_lambdified(eps, T, Eps, Sig, *args)[:, 0]

    _dDiss_dEps_lambdified = tr.Property()
    @tr.cached_property
    def _get__dDiss_dEps_lambdified(self):
        return sp.lambdify((self.eps_vars[0], self.T_var, 
                            self.Eps.as_explicit(), 
                            self.Sig.as_explicit()) + self.m_params + ('*args',), 
                           self.dDiss_dEps_, numpy_dirac, cse=True)

    dDiss_dEps_ = tr.Property()
    @tr.cached_property
    def _get_dDiss_dEps_(self):
        dFG_dEps_explicit_ = self.dF_dEps_.as_explicit()
        return (self.T_var * dFG_dEps_explicit_.diff(self.T_var) - dFG_dEps_explicit_)

    def get_dot_Eps_bounds(self, dot_eps: Any, dot_Eps: Any, *args: Any) -> Any:
        """Get bounds for internal variable rates"""
        dot_eps_sp_ = np.moveaxis(np.atleast_1d(dot_eps), -1, 0)
        dot_Eps_sp_ = np.moveaxis(dot_Eps, -1, 0)
        dot_Eps_bounds_sp = self._dot_Eps_bounds_lambdified(dot_eps_sp_, dot_Eps_sp_, *args)
        return np.moveaxis(dot_Eps_bounds_sp, 0, -1)
    
    _dot_Eps_bounds_lambdified = tr.Property()
    @tr.cached_property
    def _get__dot_Eps_bounds_lambdified(self):
        dot_eps = self.dot_eps_a[0]
        return sp.lambdify((dot_eps, 
                            self.dot_Eps.as_explicit()) + self.m_params + ('*args',), 
                           self.dot_Eps_bounds_, numpy_dirac, cse=True)

    dot_Eps_bounds_ = tr.Property()
    @tr.cached_property
    def _get_dot_Eps_bounds_(self):
        return self.dot_Eps_bounds_expr

    def get_Sig(self, eps: Any, Eps: Any, *args: Any) -> Any:
        """Calculate the stress based on the given inputs"""
        eps_sp_ = np.moveaxis(np.atleast_1d(eps), -1, 0)
        Eps_sp_ = np.moveaxis(Eps, -1, 0)
        Sig_sp = self._Sig_lambdified(eps_sp_, Eps_sp_, *args)
        Sig_sp_ = Sig_sp.reshape(Eps_sp_.shape)
        return np.moveaxis(Sig_sp_, 0, -1)
    
    _Sig_lambdified = tr.Property()
    @tr.cached_property
    def _get__Sig_lambdified(self):
        return sp.lambdify((self.eps_vars[0], 
                            self.Eps.as_explicit()) + self.m_params + ('*args',), 
                           self.Sig_.as_explicit(), numpy_dirac, cse=True)

    Sig_ = tr.Property()
    @tr.cached_property
    def _get_Sig_(self):
        return self.Y_ * self.dF_dEps_.as_explicit()

    # SymPy expression properties - minimal typing
    phi_ = tr.Property()
    @tr.cached_property  
    def _get_phi_(self):
        return (self.f_expr + self.phi_ext_expr).subs(self.m_param_subs)

    Y_ = tr.Property()
    @tr.cached_property
    def _get_Y_(self):
        Y_list = [sign_i_ * sp.eye(len(Sig_i_)) for sign_i_, Sig_i_ 
                    in zip(self.Sig_signs, self.Sig.blocks)]
        return sp.BlockDiagMatrix(*Y_list).as_explicit()

    Phi_ = tr.Property()
    @tr.cached_property
    def _get_Phi_(self):
        return -self.Y * self.phi_.diff(self.Sig.as_explicit())

    subs_Sig_Eps = tr.Property()
    @tr.cached_property
    def _get_subs_Sig_Eps(self):
        return dict(zip(self.Sig.as_explicit(), self.Sig_.as_explicit()))  # type: ignore[arg-type]

    Phi_Eps_ = tr.Property()
    @tr.cached_property
    def _get_Phi_Eps_(self):
        return self.Phi_.subs(self.subs_Sig_Eps)

    def get_Sig_f_R_dR_n1(self, eps_n: Any, d_eps: Any, Eps_n: Any, d_A: Any, d_t: Any, *args: Any) -> Tuple[Any, Any, Any, Any]:
        eps_n_sp_ = np.moveaxis(np.atleast_1d(eps_n), -1, 0)
        d_eps_sp_ = np.moveaxis(np.atleast_1d(d_eps), -1, 0)
        O_ = np.zeros_like(eps_n_sp_)
        I_ = np.ones_like(eps_n_sp_)
        d_A_sp_ = np.moveaxis(d_A, -1, 0)
        Eps_n_sp_ = np.moveaxis(Eps_n, -1, 0)
        Sig_sp_, f_sp_, R_sp_, d_R_sp_ = self._get_Sig_f_R_dR_n1_lambdified(eps_n_sp_, d_eps_sp_, Eps_n_sp_, d_A_sp_, d_t, O_, I_, *args)
        if self.phi_ == sp.S.Zero:
            f_sp_ = -np.ones_like(eps_n_sp_)
        Sig_sp_ = Sig_sp_.reshape(Eps_n_sp_.shape)
        return np.moveaxis(Sig_sp_, 0, -1), np.moveaxis(f_sp_, 0, -1), np.moveaxis(R_sp_[:, 0], 0, -1), np.moveaxis(d_R_sp_, (0, 1), (-2, -1))

    _get_Sig_f_R_dR_n1_lambdified = tr.Property()
    @tr.cached_property
    def _get__get_Sig_f_R_dR_n1_lambdified(self) -> Callable[..., Any]:
        _, (eps_n, delta_eps, Eps_n, delta_A, delta_t, self.Ox, self.Ix), Sig_n1, f_n1, R_n1, dR_dA_n1_OI = self.Sig_f_R_dR_n1
        # Convert traits tuple to regular tuple explicitly
        m_params_tuple = tuple(self.m_params)  # Force conversion
        return sp.lambdify((eps_n, delta_eps, Eps_n, delta_A, delta_t, self.Ox, self.Ix) + m_params_tuple + ('*args',), 
                            (Sig_n1, f_n1, R_n1, dR_dA_n1_OI), numpy_dirac, cse=True)
    
    Ox = sp.Symbol('O', real=True)
    Ix = sp.Symbol('I', real=True)

    def replace_zeros_and_ones_with_symbolic(self, dR_dA_: SymMatrix, delta_A: SymMatrix) -> SymMatrix:
        """Function to replace zero elements with a symbolic variable"""
        new_matrix = sp.Matrix(dR_dA_)
        rows, cols = new_matrix.shape
        for i in range(rows):
            for j in range(cols):
                if dR_dA_[i, j] == 0:
                    new_matrix[i, j] = self.Ox
                else:
                    grad_entry = dR_dA_[i, j].diff(delta_A)[:, 0]
                    ones_vector = sp.Matrix([1 for i in range(len(grad_entry))])
                    depends_on_Eps = (ones_vector.T * grad_entry)[0]
                    if depends_on_Eps == 0:
                        new_matrix[i, j] = dR_dA_[i, j] * self.Ix
                    else:
                        new_matrix[i, j] = dR_dA_[i, j]
        return new_matrix
    
    diff_along_rates = tr.Bool(True)

    Sig_f_R_dR_n1 = tr.Property()
    @tr.cached_property
    def _get_Sig_f_R_dR_n1(self) -> Tuple[Any, Any, Any, Any, Any, Any]:
        ## Manual construction of the residuum
        Eps = self.Eps.as_explicit()
        eps = self.eps_a[0]
        dot_Eps = self.dot_Eps.as_explicit()
        dot_eps = self.dot_eps_a[0]
        lam, lam_phi, f_ = self.lam_phi_f_
        dot_Lam, delta_Lam, dot_Lam_H_ = self.H_Lam

        # time symbols
        t = sp.Symbol(r't', real=True)
        delta_t = sp.Symbol(r'\Delta t', real=True)

        # fundamental state
        Eps_n = sp.Matrix([sp.Symbol(f'{var.name}_{{(n)}}', real=True) for var in Eps])
        eps_n = sp.Symbol(r'\varepsilon_{(n)}', real=True)

        # increment
        delta_Eps = sp.Matrix([sp.Symbol(f'\\Delta{{{var.name}}}', real=True) for var in Eps])
        delta_eps = sp.Symbol(r'\Delta{\varepsilon}', real=True)
        delta_lam = sp.Matrix([sp.Symbol(f'\\Delta{{{var.name}}}', real=True) for var in lam])
        
        # updated state
        Eps_n1 = Eps_n + delta_Eps
        eps_n1 = eps_n + delta_eps

        # rate of change
        dot_Eps_n = delta_Eps / delta_t
        dot_eps_n = delta_eps / delta_t
        dot_lam_n = delta_lam / delta_t


        # derive substitutions
        subs_dot_Eps = dict(zip(dot_Eps, dot_Eps_n))  # type: ignore[arg-type]
        subs_dot_eps = {dot_eps: dot_eps_n}
        subs_delta_lam = dict(zip(lam, delta_lam))  # type: ignore[arg-type]
        subs_dot_Lam = dict(zip(dot_Lam, delta_Lam))  # type: ignore[arg-type]
        subs_Eps_n1 = dict(zip(Eps, Eps_n1))  # type: ignore[arg-type]
        subs_eps_n1 = {eps: eps_n1}
        subs_dt = {self.d_t: delta_t}

        subs_n1 = {**subs_dot_Eps, **subs_dot_eps, **subs_delta_lam, 
                   **subs_dot_Lam, **subs_Eps_n1, **subs_eps_n1, **subs_dt}

        Sig = self.Sig.as_explicit()
        Sig_ = self.Sig_.as_explicit()

        f_ = self.f_expr

        gamma_mech = self.gamma_mech_sign * ((self.Y * Sig).T * self.dot_Eps.as_explicit())[0]

        # Full Lagrangian for the minimum principle of dissipation potential
        L_ = -gamma_mech - dot_Lam_H_ - lam_phi 

        # Generalized forces and flux increments
        S = sp.Matrix([Sig, sp.Matrix(dot_Lam), sp.Matrix(lam)])
        delta_A = sp.Matrix([delta_Eps, sp.Matrix(delta_Lam), sp.Matrix(delta_lam)])

        # Derivative of the Lagrangian with respect to the generalized forces
        dL_dS_ = L_.diff(S)
        dL_dS_ = sp.Matrix(dL_dS_)
        if self.f_expr != sp.S.Zero:
            dL_dS_[-1] = f_

        dL_dS_A_ = dL_dS_.subs(self.subs_Sig_Eps).subs(self.subs_sig_eps)
        R_n1 = dL_dS_A_.subs(subs_n1)

        # Jacobian of the residuum
        dR_dA_n1 = R_n1.jacobian(delta_A)
        dR_dA_n1 = dR_dA_n1.replace(sp.Derivative, lambda *args: 0)
        dR_dA_n1 = dR_dA_n1.replace(sp.DiracDelta, lambda *args: 0)

        # replace zeros and constant terms with symbolic variables to broadcast properly
        dR_dA_n1_OI = self.replace_zeros_and_ones_with_symbolic(dR_dA_n1, delta_A)

        # Values of the static threshold function
        f_Eps_ = f_.subs(self.subs_Sig_Eps)
        f_n1 = f_Eps_.subs(subs_n1)

        # External stress
        Sig_n1 = Sig_.subs(subs_n1)

        return (gamma_mech, L_, dL_dS_, dL_dS_A_, dR_dA_n1), (eps_n, delta_eps, Eps_n, delta_A, delta_t, self.Ox, self.Ix), Sig_n1, f_n1, R_n1, dR_dA_n1_OI

    def get_state_n1(self, eps_n: Any, d_eps: Any, d_t: Any, Eps_n: Any, *args: Any) -> Tuple[Any, Any, Any, Any]:
        """Calculate the state at time n+1 based on the given inputs using an iterative algorithm"""
        n_I = np.atleast_1d(eps_n).shape[0]
        tol = 1e-8
        k_I = np.zeros((n_I,), dtype=np.int_)
        d_A = np.zeros((n_I, self.n_Eps + self.n_Lam + self.n_lam), dtype=np.float64)
        
        Sig_n1, f_n1, R_n1, dR_dA_n1 = self.get_Sig_f_R_dR_n1(
            eps_n, d_eps, Eps_n, d_A, d_t, *args)

        I = f_n1 >= 0
        I_inel = np.copy(I)
        I_el = ~I_inel

        # Inelastic state update - only when inequality constraint is present
        if self.n_lam > 0:
            for k in range(self.k_max):
                if np.all(I == False):
                    break
                try:
                    d_A[I] -= np.linalg.solve(dR_dA_n1[I], R_n1[I][..., np.newaxis])[..., 0]
                except np.linalg.LinAlgError as e:
                    print("SingularMatrix encountered in dR_dA_n1[I]:", dR_dA_n1[I])
                    print(f"eps = {eps_n}, d_eps = {d_eps}, Eps_n = {Eps_n}, d_A = {d_A}, d_t = {d_t}")
                    raise
                
                Sig_n1[I], f_n1[I], R_n1[I], dR_dA_n1[I] = self.get_Sig_f_R_dR_n1(eps_n[I], d_eps[I], Eps_n[I], d_A[I], d_t, *args)
                norm_R_n1 = np.linalg.norm(R_n1, axis=-1)
                I[norm_R_n1 <= tol] = False
                k_I[I] += 1

            # If internal variables are out of bounds set the inelastic state to elastic
            if self.dot_Eps_bounds_expr != sp.S.Zero:
                I_el[self.get_dot_Eps_bounds(d_eps, d_A[..., :self.n_Eps], *args) > 0] = True

        # Elastic state update
        if self.n_Lam > 0:
            for k in range(self.k_max):
                if np.all(I_el == False):
                    break
                try:
                    if self.n_lam == 0:
                        i1 = None
                    else:
                        i1 = -1
                        d_A[I_el, i1] = 0
                    d_A[I_el, :i1] -= np.linalg.solve(dR_dA_n1[I_el, :i1, :i1], R_n1[I_el, :i1, np.newaxis])[..., 0]
                except np.linalg.LinAlgError as e:
                    print("SingularMatrix encountered in dR_dA_n1[I]:", dR_dA_n1[I_el])
                    print(f"eps = {eps_n}, d_eps = {d_eps}, Eps_n = {Eps_n}, d_A = {d_A}, d_t = {d_t}")
                    raise
                
                Sig_n1[I_el], f_n1[I_el], R_n1[I_el], dR_dA_n1[I_el] = self.get_Sig_f_R_dR_n1(eps_n[I_el], d_eps[I_el], Eps_n[I_el], d_A[I_el], d_t, *args)
                norm_R_n1 = np.linalg.norm(R_n1[...,:i1], axis=-1)
                k_I[I_el] += 1
                I_el[norm_R_n1 <= tol] = False

        lam_k = d_A[..., self.n_Eps:]
        Eps_n1 = Eps_n + d_A[..., :self.n_Eps]

        return Eps_n1, Sig_n1, lam_k, k_I

    def get_response(self, eps_ta: npt.NDArray[np.float64], t_t: npt.NDArray[np.float64], *args: float) -> Tuple[npt.NDArray[np.float64], ...]:
        """Time integration procedure"""
        if eps_ta.ndim == 2:
            eps_ta = eps_ta[:,np.newaxis,:]
        if eps_ta.ndim == 1:
            eps_ta = eps_ta[:, np.newaxis]

        d_eps_ta = np.diff(eps_ta, axis=0)
        d_t_t = np.diff(t_t, axis=0)
        Eps_n1 = np.zeros(eps_ta.shape[1:] + (self.n_Eps,), dtype=np.float64)
        Sig_n1 = np.zeros_like(Eps_n1)
        lam_n1 = np.zeros(eps_ta.shape[1:] + (self.n_lam + self.n_Lam,), dtype=np.float64)
        lam_n1 = np.zeros(eps_ta.shape[1:] + (self.n_lam + self.n_Lam,), dtype=np.float64)
        Sig_record = [Sig_n1]
        Eps_record = [Eps_n1]
        iter_record = [np.zeros(eps_ta.shape[1:])]
        lam_record = [lam_n1]
        for n, dt in enumerate(d_t_t):
            print('increment', n+1, end='\r')
            try:
                Eps_n1, Sig_n1, lam_n1, k = self.get_state_n1(
                    eps_ta[n], d_eps_ta[n], dt, Eps_n1, *args
                )
            except RuntimeError as e:
                print(f'{n+1}({k}) ... {str(e)}', end='\r')
                break
            Sig_record.append(Sig_n1)
            Eps_record.append(Eps_n1)
            lam_record.append(lam_n1)
            iter_record.append(k)
        Sig_t = np.array(Sig_record, dtype=np.float64)
        Eps_t = np.array(Eps_record, dtype=np.float64)
        iter_t = np.array(iter_record,dtype=np.int_)
        lam_t = np.array(lam_record,dtype=np.float64)
        n_t = len(Eps_t)
        eps_ta = eps_ta[:n_t]
        sig_ta = self.get_sig_a(eps_ta[..., np.newaxis], Eps_t, *args)
        return (t_t[:n_t], eps_ta, sig_ta, Eps_t, Sig_t, iter_t, lam_t, (d_t_t[:n_t], d_eps_ta[:n_t]))

    def save_to_disk(self) -> None:
        """Serialize this instance to disk."""
        filepath = os.path.join(CACHE_DIR, f"{self.name}.pkl")
        os.makedirs(CACHE_DIR, exist_ok=True)
        with open(filepath, 'wb') as file:
            dill.dump(self, file)

    @staticmethod
    def load_from_disk(name: str) -> 'GSMEngine':
        """Deserialize an instance from disk."""
        filepath = os.path.join(CACHE_DIR, f"{name}.pkl")
        if not os.path.exists(filepath):
            raise FileNotFoundError(f"No cached data found for {name}.")
        with open(filepath, 'rb') as file:
            return cast('GSMEngine', dill.load(file))



"""
Pure GSMSymbDef Implementation - Phase 1

This module provides a pure symbolic thermodynamic definition class that separates
symbolic expressions from numerical execution engines.

Key Features:
- Contains only symbolic thermodynamic definitions
- No embedded engines
- Supports model specialization
- Provides expression validation and rendering
- Includes codename mappings for user-friendly display
"""

import sympy as sp
import keyword
from typing import Dict, Tuple, Any, Optional, List, NamedTuple
from dataclasses import dataclass
from functools import cached_property


def is_valid_variable_name(name: str) -> bool:
    """Check if the given name is a valid Python variable name."""
    if not name.isidentifier():
        return False
    if keyword.iskeyword(name):
        return False
    return True


@dataclass
class ValidationResult:
    """Result of symbolic expression validation."""
    is_valid: bool
    issues: List[str]


class GSMSymbDef:
    """
    Pure symbolic thermodynamic definition for GSM framework.
    
    This class contains only symbolic expressions and variable definitions
    without any embedded numerical execution engines. This is the non-discretized
    part of the model formulation.
    """
    
    # Primary variables (always Helmholtz-based)
    eps_vars: Tuple[sp.Symbol, ...] = ()      # External strain variables (ε)
    sig_vars: Tuple[sp.Symbol, ...] = ()      # External stress variables (σ) - conjugate to eps_vars
    Eps_vars: Tuple[sp.Symbol, ...] = ()      # Internal strain variables (ℰ)  
    Sig_vars: Tuple[sp.Symbol, ...] = ()      # Thermodynamic force variables (𝒮) - conjugate to Eps_vars
    m_params: Tuple[sp.Symbol, ...] = ()      # Material parameters
    
    # Thermodynamic expressions
    F_expr: sp.Expr = sp.S.Zero               # Free energy (state potential)
    f_expr: sp.Expr = sp.S.Zero               # Dissipation potential (inequality constraint)
    phi_ext_expr: sp.Expr = sp.S.Zero         # External potential
    h_k: List[sp.Expr] = []                   # Equality constraints
    
    # Variable properties
    Sig_signs: Tuple[int, ...] = ()           # Signs for internal stress variables (𝒮)
    gamma_mech_sign: int = -1                 # Sign for mechanical dissipation (Helmholtz: -1, Gibbs: +1)
    
    # Display mappings
    eps_codenames: Dict[sp.Symbol, str] = {}   # Strain variable codenames (ε)
    sig_codenames: Dict[sp.Symbol, str] = {}   # Stress variable codenames (σ)
    Eps_codenames: Dict[sp.Symbol, str] = {}   # Internal strain codenames (ℰ)
    Sig_codenames: Dict[sp.Symbol, str] = {}   # Thermodynamic force codenames (𝒮)
    param_codenames: Dict[sp.Symbol, str] = {} # Parameter codenames

    def __init__(self):
        """Initialize GSMSymbDef instance."""
        # Initialize empty constraint list if not defined
        if not hasattr(self, 'h_k'):
            self.h_k = []
            
        # Validate the model on initialization
        validation = self.validate_symbolic_expressions()
        if not validation.is_valid:
            print(f"⚠️ Warning: Model validation issues in {self.__class__.__name__}:")
            for issue in validation.issues:
                print(f"  - {issue}")

    @classmethod
    def get_required_parameters(cls) -> List[str]:
        """
        Return list of required parameter codenames.
        
        Returns:
            List of parameter names (codenames) required for this model
        """
        param_names = []
        for param in cls.m_params:
            if param in cls.param_codenames:
                param_names.append(cls.param_codenames[param])
            elif is_valid_variable_name(param.name):
                param_names.append(param.name)
            else:
                param_names.append(str(param))
        return param_names
    
    @classmethod  
    def get_parameter_descriptions(cls) -> Dict[str, str]:
        """
        Return parameter descriptions for documentation.
        
        Returns:
            Dictionary mapping parameter names to descriptions
        """
        descriptions = {}
        for param in cls.m_params:
            param_name = cls.param_codenames.get(param, param.name)
            # Extract description from symbol assumptions or provide default
            if hasattr(param, '_description'):
                descriptions[param_name] = param._description
            else:
                # Generate basic description from symbol properties
                desc_parts = []
                if param.is_positive:
                    desc_parts.append("positive")
                if param.is_real:
                    desc_parts.append("real")
                if desc_parts:
                    descriptions[param_name] = f"Material parameter ({', '.join(desc_parts)})"
                else:
                    descriptions[param_name] = "Material parameter"
        return descriptions
    
    @classmethod
    def validate_symbolic_expressions(cls) -> ValidationResult:
        """
        Validate consistency of symbolic expressions.
        
        Returns:
            ValidationResult indicating if model is valid and any issues found
        """
        issues = []
        
        # Check if expressions are defined
        if cls.F_expr == sp.S.Zero:
            issues.append("F_expr (free energy) is not defined")
            
        # Get all symbols used in expressions
        F_symbols = cls.F_expr.free_symbols if cls.F_expr != sp.S.Zero else set()
        f_symbols = cls.f_expr.free_symbols if cls.f_expr != sp.S.Zero else set()
        phi_symbols = cls.phi_ext_expr.free_symbols if cls.phi_ext_expr != sp.S.Zero else set()
        
        # Collect constraint symbols
        h_symbols = set()
        for constraint in cls.h_k:
            h_symbols.update(constraint.free_symbols)
        
        # All defined symbols
        all_defined_symbols = set(cls.eps_vars + cls.sig_vars + cls.Eps_vars + cls.Sig_vars + cls.m_params)
        
        # Check for undefined symbols
        all_used_symbols = F_symbols | f_symbols | phi_symbols | h_symbols
        undefined_symbols = all_used_symbols - all_defined_symbols
        
        if undefined_symbols:
            issues.append(f"Undefined symbols: {undefined_symbols}")
            
        # Check codename completeness
        missing_codenames = []
        for symbol_group, codename_dict, group_name in [
            (cls.eps_vars, cls.eps_codenames, "strain variables"),
            (cls.sig_vars, cls.sig_codenames, "stress variables"),
            (cls.Eps_vars, cls.Eps_codenames, "internal variables"),
            (cls.Sig_vars, cls.Sig_codenames, "thermodynamic forces"),
            (cls.m_params, cls.param_codenames, "parameters")
        ]:
            for symbol in symbol_group:
                if symbol not in codename_dict and not is_valid_variable_name(symbol.name):
                    missing_codenames.append(f"{symbol} in {group_name}")
        
        if missing_codenames:
            issues.append(f"Missing codenames for: {', '.join(missing_codenames)}")
        
        # Check consistency between eps_vars and sig_vars
        if len(cls.eps_vars) != len(cls.sig_vars):
            issues.append(f"Mismatch: {len(cls.eps_vars)} external strain variables but {len(cls.sig_vars)} external stress variables")
            
        # Check sign definitions for internal variables (only if Sig_signs is defined)
        if hasattr(cls, 'Sig_signs') and cls.Sig_signs:
            if len(cls.Eps_vars) != len(cls.Sig_signs):
                issues.append(f"Mismatch: {len(cls.Eps_vars)} internal variables but {len(cls.Sig_signs)} signs")
        else:
            if cls.Eps_vars:  # Only warn if there are internal variables
                issues.append("Sig_signs not defined for internal variables")
            
        # Check consistency between Eps_vars and Sig_vars (only if both are defined)
        if cls.Eps_vars and cls.Sig_vars:
            if len(cls.Eps_vars) != len(cls.Sig_vars):
                issues.append(f"Mismatch: {len(cls.Eps_vars)} internal variables but {len(cls.Sig_vars)} thermodynamic forces")
        elif cls.Eps_vars and not cls.Sig_vars:
            issues.append(f"Mismatch: {len(cls.Eps_vars)} internal variables but 0 thermodynamic forces")
        
        return ValidationResult(is_valid=len(issues) == 0, issues=issues)

    def render_potentials(self) -> str:
        """
        Render thermodynamic potentials in LaTeX format.
        
        Returns:
            LaTeX string representation of the model potentials
        """
        latex_lines = []
        
        # Model header
        latex_lines.append(f"## {self.__class__.__name__} - Thermodynamic Potentials")
        latex_lines.append("")
        
        # Free energy (Helmholtz)
        latex_lines.append("### State Potential (Helmholtz Free Energy)")
        if self.F_expr != sp.S.Zero:
            latex_lines.append(f"$$F = {sp.latex(self.F_expr)}$$")
        else:
            latex_lines.append("$$F = 0$$ (not defined)")
        latex_lines.append("")
        
        # Dissipation potential
        latex_lines.append("### Dissipation Potential (Inequality Constraint)")
        if self.f_expr != sp.S.Zero:
            latex_lines.append(f"$$f = {sp.latex(self.f_expr)} \\leq 0$$")
        else:
            latex_lines.append("$$f = 0$$ (no dissipation)")
        latex_lines.append("")
        
        # External potential
        if self.phi_ext_expr != sp.S.Zero:
            latex_lines.append("### External Potential")
            latex_lines.append(f"$$\\phi_{{ext}} = {sp.latex(self.phi_ext_expr)}$$")
            latex_lines.append("")
        
        # Equality constraints
        if self.h_k:
            latex_lines.append("### Equality Constraints")
            for i, constraint in enumerate(self.h_k):
                latex_lines.append(f"$$h_{{{i}}} = {sp.latex(constraint)} = 0$$")
            latex_lines.append("")
        
        return "\n".join(latex_lines)
    
    def get_expression_summary(self) -> Dict[str, Any]:
        """
        Get summary of all symbolic expressions.
        
        Returns:
            Dictionary containing model summary information
        """
        return {
            "model_name": self.__class__.__name__,
            "external_variables": len(self.eps_vars),
            "external_stresses": len(self.sig_vars),
            "internal_variables": len(self.Eps_vars),
            "thermodynamic_forces": len(self.Sig_vars),
            "material_parameters": len(self.m_params),
            "has_free_energy": self.F_expr != sp.S.Zero,
            "has_dissipation": self.f_expr != sp.S.Zero,
            "has_external_potential": self.phi_ext_expr != sp.S.Zero,
            "equality_constraints": len(self.h_k),
            "free_energy_complexity": len(self.F_expr.free_symbols) if self.F_expr != sp.S.Zero else 0,
            "dissipation_complexity": len(self.f_expr.free_symbols) if self.f_expr != sp.S.Zero else 0
        }

    # Derived expressions as cached properties
    @cached_property
    def eps_a(self) -> sp.Matrix:
        """
        External strain variables as a vector (eps_a).
        
        Following the convention where _a index represents the expanded collection
        of all external control strain variables as a homogeneous vector structure.
        
        Returns:
            Vector (Matrix) of external strain variables
        """
        return sp.BlockMatrix([
            var if isinstance(var, sp.Matrix) and var.shape[1] == 1 else sp.Matrix([[var]]) 
            for var in self.eps_vars
        ]).as_explicit()

    @cached_property
    def sig_a(self) -> sp.Matrix:
        """
        External stress variables as a vector (sig_a).
        
        Following the convention where _a index represents the expanded collection
        of all external stress variables as a homogeneous vector structure.
        These are the conjugate variables to eps_a.
        
        Returns:
            Vector (Matrix) of external stress variables
        """
        return sp.BlockMatrix([
            var if isinstance(var, sp.Matrix) and var.shape[1] == 1 else sp.Matrix([[var]]) 
            for var in self.sig_vars
        ]).as_explicit()

    @cached_property
    def Eps_B(self) -> sp.Matrix:
        """
        Internal strain variables with global indexing for heterogeneous types (Eps_B).
        
        Following the gsm_engine.py convention where individual multi-dimensional internal 
        variables (scalar, vector, tensor) are organized in a BlockMatrix structure with 
        global indexing scheme. The _B suffix (capital) indicates the block-structured 
        representation with non-expanded heterogeneous variable types (e.g., plastic tensor, 
        isotropic hardening scalar, kinematic hardening tensor).
        
        Similar to gsm_engine.py's Eps property: BlockMatrix([Eps_i.T for Eps_i in self.Eps_vars]).T
        
        Returns:
            BlockMatrix of internal strain variables maintaining their original dimensions
        """
        if not self.Eps_vars:
            return sp.Matrix([])
        
        # Create list of internal variables transposed for BlockMatrix structure
        # Following gsm_engine.py pattern: [Eps_i.T for Eps_i in self.Eps_vars]
        Eps_list_transposed = [
            var.T if isinstance(var, sp.Matrix) 
            else sp.Matrix([[var]]) 
            for var in self.Eps_vars
        ]
        
        # Use BlockMatrix and transpose to match gsm_engine.py: BlockMatrix(...).T
        return sp.BlockMatrix(Eps_list_transposed).T

    @cached_property
    def Eps_b(self) -> sp.Matrix:
        """
        Internal strain variables as a flattened vector (Eps_b).
        
        Following the convention where _b index (lowercase) represents the expanded/flattened 
        collection of all internal strain variables as a homogeneous vector structure.
        The individual variables from Eps_B are converted from their semantically 
        distinguished form (Scalar, Vector, Tensor) to flat form with expanded indexing.
        
        This is derived from Eps_B using as_explicit() to flatten the block structure.
        
        Returns:
            Flattened vector (Matrix) of all internal strain variable components
        """
        return self.Eps_B.as_explicit()

    @cached_property
    def Sig_B(self) -> sp.Matrix:
        """
        Thermodynamic force variables with global indexing for heterogeneous types (Sig_B).
        
        Following the gsm_engine.py convention where individual multi-dimensional thermodynamic
        force variables (scalar, vector, tensor) are organized in a BlockMatrix structure with 
        global indexing scheme. The _B suffix (capital) indicates the block-structured 
        representation with non-expanded heterogeneous variable types.
        
        Returns:
            BlockMatrix of thermodynamic force variables maintaining their original dimensions
        """
        if not self.Sig_vars:
            return sp.Matrix([])
        
        # Create list of thermodynamic force variables transposed for BlockMatrix structure
        Sig_list_transposed = [
            var.T if isinstance(var, sp.Matrix) 
            else sp.Matrix([[var]]) 
            for var in self.Sig_vars
        ]
        
        # Use BlockMatrix and transpose to match gsm_engine.py pattern
        return sp.BlockMatrix(Sig_list_transposed).T

    @cached_property
    def Sig_b(self) -> sp.Matrix:
        """
        Thermodynamic force variables as a flattened vector (Sig_b).
        
        Following the convention where _b index (lowercase) represents the expanded/flattened 
        collection of all thermodynamic force variables as a homogeneous vector structure.
        The individual variables from Sig_B are converted from their semantically 
        distinguished form (Scalar, Vector, Tensor) to flat form with expanded indexing.
        
        This is derived from Sig_B using as_explicit() to flatten the block structure.
        
        Returns:
            Flattened vector (Matrix) of all thermodynamic force variable components
        """
        return self.Sig_B.as_explicit()

    @cached_property
    def dot_eps_a(self) -> sp.Matrix:
        """
        Rate of external strain variables (dot_eps_a).
        
        Creates time derivatives of all external strain variables following the
        gsm_engine.py pattern. Each variable gets a dot symbol prefix.
        
        Returns:
            Vector (Matrix) of external strain rate symbols
        """
        return sp.Matrix([sp.Symbol(f'\\dot{{{var.name}}}', real=True) for var in self.eps_a])

    @cached_property
    def dot_Eps_b(self) -> sp.Matrix:
        """
        Rate of internal strain variables (dot_Eps_b).
        
        Creates time derivatives of all internal strain variables in flattened form,
        following the gsm_engine.py pattern. Each variable gets a dot symbol prefix.
        
        Returns:
            Vector (Matrix) of internal strain rate symbols (flattened form)
        """
        return sp.Matrix([sp.Symbol(f'\\dot{{{var.name}}}', real=True) for var in self.Eps_b])

    @cached_property
    def sig_a_(self) -> sp.Matrix:
        """
        External stresses obtained as derivatives of state function w.r.t. external strains (expression).
        
        General formulation: σ = ∂F/∂ε (sign determined by flow variable configuration)
        Following the convention where _a index represents the expanded collection.
        The trailing underscore indicates this is a derived expression, not a symbol.
        
        Returns:
            Vector (Matrix) of stress expressions (σ) corresponding to eps_a
        """
        if self.F_expr == sp.S.Zero:
            return sp.Matrix([sp.S.Zero for _ in range(len(self.eps_a))])
        
        return self.F_expr.diff(self.eps_a)

    @cached_property
    def dF_dEps_b(self) -> sp.Matrix:
        """
        Derivatives of free energy with respect to internal strain variables.
        
        Direct differentiation w.r.t. flattened Eps_b to get vector of derivatives.
        
        Returns:
            Vector of derivatives ∂F/∂ℰ_b (expanded/flattened form)
        """
        if self.F_expr == sp.S.Zero or not self.Eps_vars:
            return sp.Matrix([])
        
        return sp.Matrix([sp.diff(self.F_expr, var) for var in self.Eps_b])

    @cached_property
    def Y_b(self) -> sp.Matrix:
        """
        Sign matrix for thermodynamic forces (expanded representation).
        
        Expands the block-wise signs in Sig_signs to match the flattened Eps_b structure.
        Creates a diagonal matrix with appropriate signs for each component.
        
        Returns:
            Diagonal matrix of signs for expanded internal variables
        """
        if not self.Eps_vars:
            return sp.Matrix([])
        
        # Expand signs to match flattened structure
        signs_expanded = []
        for i, var in enumerate(self.Eps_vars):
            sign = self.Sig_signs[i] # if i < len(cls.Sig_signs) else -1
            # Get the size of this variable when flattened
            if isinstance(var, sp.Matrix):
                var_size = var.shape[0] * var.shape[1]
            else:
                var_size = 1
            signs_expanded.extend([sign] * var_size)
        
        return sp.diag(*signs_expanded)

    @cached_property
    def Sig_b_(self) -> sp.Matrix:
        """
        Thermodynamic forces obtained from free energy derivatives with sign convention (expression).
        
        General formulation: 𝒮 = Y * ∂F/∂ℰ (sign convention configures state domain)
        Following the gsm_engine.py convention: Sig_ = Y_ * dF_dEps_.as_explicit()
        The trailing underscore indicates this is a derived expression, not a symbol.
        
        Returns:
            Vector (Matrix) of thermodynamic force expressions (𝒮) corresponding to Eps_b
        """
        if self.F_expr == sp.S.Zero or not self.Eps_vars:
            return sp.Matrix([])
        
        return self.Y_b * self.dF_dEps_b

    @cached_property
    def phi_(self) -> sp.Expr:
        """
        Combined dissipation and flow potential (phi_).
        
        Following the gsm_engine.py convention where phi_ combines the threshold 
        function (f_expr) and the flow potential extension (phi_ext_expr).
        This represents the complete dissipation criterion for the material model.
        
        In gsm_engine.py: phi_ = (f_expr + phi_ext_expr).subs(m_param_subs)
        Since m_param_subs is not used in practice, we omit the substitution.
        
        Returns:
            Combined dissipation potential expression
        """
        return self.f_expr + self.phi_ext_expr

    @cached_property
    def subs_sig_eps_a(self) -> Dict[sp.Symbol, sp.Expr]:
        """
        Constitutive relations for external variables: sig_a = sig_a_
        
        Creates substitution dictionary mapping external stress symbols to
        their expressions derived from free energy derivatives.
        
        Returns:
            Dictionary {sigma_i: ∂F/∂ε_i} for external constitutive relations
        """
        return dict(zip(self.sig_a, self.sig_a_))

    @cached_property  
    def subs_Sig_Eps_b(self) -> Dict[sp.Symbol, sp.Expr]:
        """
        Constitutive relations for internal variables: Sig_b = Sig_b_
        
        Creates substitution dictionary mapping thermodynamic force symbols to
        their expressions derived from free energy derivatives with sign convention.
        
        Returns:
            Dictionary {S_i: Y_b * ∂F/∂ℰ_i} for internal constitutive relations
        """
        return dict(zip(self.Sig_b, self.Sig_b_))

    @cached_property
    def gamma_mech_(self) -> sp.Expr:
        """
        Mechanical dissipation (gamma_mech_).
        
        Following the gsm_engine.py convention where mechanical dissipation is calculated as:
        gamma_mech = gamma_mech_sign * ((Y * Sig).T * dot_Eps)[0]
        
        This represents the power dissipated through internal variable evolution.
        The sign depends on the thermodynamic potential type:
        - Helmholtz (gamma_mech_sign = -1): negative dissipation
        - Gibbs (gamma_mech_sign = +1): positive dissipation
        
        Returns:
            Mechanical dissipation expression
        """
        if not self.Eps_vars:
            return sp.S.Zero
        
        # Calculate mechanical dissipation: gamma_mech_sign * ((Y_b * Sig_b).T * dot_Eps_b)[0]
        Y_Sig_product = self.Y_b * self.Sig_b
        dissipation_rate = (Y_Sig_product.T * self.dot_Eps_b)[0]
        
        return self.gamma_mech_sign * dissipation_rate

    @cached_property
    def lam_phi_f_(self) -> Tuple[List[sp.Symbol], sp.Expr, sp.Expr]:
        """
        Inequality constraint components: Lagrange multipliers, constraint term, and flow potential.
        
        Following gsm_engine.py pattern:
        lam_phi_f_ = (lam, lam_phi, f_)
        where lam is list of multipliers, lam_phi is constraint term, f_ is flow potential
        
        Returns:
            Tuple of (multiplier_list, constraint_term, flow_potential)
        """
        if self.phi_ == sp.S.Zero:
            return ([], sp.S.Zero, sp.S.Zero)
        
        lam_phi = sp.Symbol(r'\lambda_{\mathrm{\phi}}', real=True)
        return ([lam_phi], lam_phi * (-self.phi_), self.f_expr)

    @cached_property
    def H_Lam_(self) -> Tuple[List[sp.Symbol], sp.Matrix, sp.Expr]:
        """
        Equality constraint components: rate multipliers, increment multipliers, and constraint sum.
        
        Following gsm_engine.py pattern:
        H_Lam = (dot_Lam, delta_Lam, dot_lam_sum)
        where dot_Lam are rate multipliers, delta_Lam are incremental multipliers, 
        dot_lam_sum is the constraint term sum
        
        Returns:
            Tuple of (rate_multipliers, increment_multipliers, constraint_sum)
        """
        dot_Lam = [sp.Symbol(f'\\dot{{\\lambda}}_{{{k}}}', real=True) for k in range(len(self.h_k))]
        delta_Lam = sp.Matrix([sp.Symbol(f'\\Delta\\lambda_{{{k}}}', real=True) for k in range(len(self.h_k))])
        dot_lam_sum = sum(l * h for l, h in zip(dot_Lam, self.h_k)) if self.h_k else sp.S.Zero
        return (dot_Lam, delta_Lam, dot_lam_sum)

    @cached_property
    def L_(self) -> sp.Expr:
        """
        Lagrangian for the minimum principle of dissipation potential.
        
        Following gsm_engine.py convention:
        L_ = -gamma_mech - dot_Lam_H_ - lam_phi
        
        This represents the full Lagrangian combining:
        - Mechanical dissipation (negative for energy balance)
        - Equality constraint terms (from h_k constraints)
        - Inequality constraint terms (from flow potential phi_)
        
        Returns:
            Complete Lagrangian expression
        """
        # Get constraint components
        lam_phi_list, lam_phi_term, _ = self.lam_phi_f_
        dot_Lam_list, delta_Lam, dot_lam_sum = self.H_Lam_
        
        # Construct full Lagrangian: L_ = -gamma_mech - dot_Lam_H_ - lam_phi
        L_expr = -self.gamma_mech_ - dot_lam_sum - lam_phi_term
        
        return L_expr

    @cached_property
    def dL_dS_A_(self) -> sp.Matrix:
        """
        Continuous optimality criterion: derivative of Lagrangian w.r.t. generalized forces.
        
        Following gsm_engine.py convention where dL_dS_A_ represents the derivative of the
        Lagrangian with respect to the generalized forces after applying constitutive
        relations. This should be zero for the optimal state with maximum energy dissipation.
        
        The generalized forces vector S includes:
        - Internal thermodynamic forces (Sig_b)
        - Lagrange multipliers for equality constraints
        - Lagrange multipliers for inequality constraints
        
        Returns:
            Matrix of optimality conditions that should equal zero
        """
        if not self.Eps_vars:
            return sp.Matrix([])
        
        # Get constraint components
        lam_phi_list, lam_phi_term, f_constraint = self.lam_phi_f_
        dot_Lam_list, delta_Lam, dot_lam_sum = self.H_Lam_
        
        # Construct generalized forces vector S = [Sig_b, dot_Lam, lam_phi]
        S_components = [self.Sig_b]
        if dot_Lam_list:
            S_components.append(sp.Matrix(dot_Lam_list))
        if lam_phi_list:
            S_components.append(sp.Matrix(lam_phi_list))
        
        if len(S_components) == 1:
            S = S_components[0]
        else:
            S = sp.Matrix.vstack(*S_components)
        
        # Calculate derivative of Lagrangian with respect to generalized forces
        dL_dS = self.L_.diff(S)
        dL_dS_matrix = sp.Matrix(dL_dS) if not isinstance(dL_dS, sp.Matrix) else dL_dS
        
        # Apply constitutive relations (substitute Sig_b expressions)
        if self.subs_Sig_Eps_b and self.subs_sig_eps_a:
            # Combine both external and internal constitutive relations
            all_subs = {**self.subs_Sig_Eps_b, **self.subs_sig_eps_a}
            dL_dS_A = dL_dS_matrix.subs(all_subs)
        elif self.subs_Sig_Eps_b:
            dL_dS_A = dL_dS_matrix.subs(self.subs_Sig_Eps_b)
        else:
            dL_dS_A = dL_dS_matrix
        
        # Handle the special case for inequality constraint (flow potential)
        if lam_phi_list and self.f_expr != sp.S.Zero:
            # The last component should be the flow potential constraint f_
            dL_dS_A[-1] = self.f_expr
        
        return dL_dS_A

    @cached_property
    def subs_n1_(self) -> Dict[sp.Symbol, sp.Expr]:
        """
        Time discretization substitutions for state variables at time n+1.
        
        Following gsm_engine.py pattern from _get_Sig_f_R_dR_n1 method where
        the continuous time derivatives and state variables are replaced with
        discrete incremental forms for numerical time integration.
        
        Creates substitutions for:
        - Rate variables: dot_Eps, dot_eps, dot_Lam -> increments/delta_t
        - State variables: Eps, eps -> n-state + increments
        - Lagrange multipliers: lam -> delta_lam
        - Time differential: d_t -> delta_t
        
        Returns:
            Dictionary of substitutions for time discretization
        """
        from bmcs_matmod.gsm_lagrange.core.gsm_vars import Scalar
        
        # Time symbols
        delta_t = Scalar(r'\Delta t', codename='delta_t', real=True)
        
        # External variables - state at time n (previous time step)
        eps_n = sp.Matrix([
            Scalar(f'{var.name}_{{(n)}}', codename=f'{self._get_codename(var)}_n', real=True) 
            for var in self.eps_a
        ])
        
        # Internal variables - state at time n (previous time step)
        Eps_n = sp.Matrix([
            Scalar(f'{var.name}_{{(n)}}', codename=f'{self._get_codename(var)}_n', real=True) 
            for var in self.Eps_b
        ])
        delta_Eps = sp.Matrix([
            Scalar(f'\\Delta{{{var.name}}}', codename=f'delta_{self._get_codename(var)}', real=True) 
            for var in self.Eps_b
        ])
        
        # Updated state at time n+1
        eps_n1 = eps_n + self.delta_eps_a_
        Eps_n1 = Eps_n + delta_Eps
        
        # Rate of change (increment / time_step)
        dot_eps_n = self.delta_eps_a_ / delta_t
        dot_Eps_n = delta_Eps / delta_t
        
        # Build substitution dictionary
        subs_dict = {}
        
        # External strain substitutions - iterate over all components
        for i, (eps_var, dot_eps_var) in enumerate(zip(self.eps_a, self.dot_eps_a)):
            subs_dict[eps_var] = eps_n1[i]
            subs_dict[dot_eps_var] = dot_eps_n[i]
        
        # Internal strain substitutions
        for i, (eps_var, dot_eps_var) in enumerate(zip(self.Eps_b, self.dot_Eps_b)):
            subs_dict[eps_var] = Eps_n1[i]
            subs_dict[dot_eps_var] = dot_Eps_n[i]
        
        # Constraint multiplier substitutions
        lam_phi_list, lam_phi_term, _ = self.lam_phi_f_
        dot_Lam_list, delta_Lam, dot_lam_sum = self.H_Lam_
        
        # Equality constraint Lagrange multipliers
        for k, dot_lam in enumerate(dot_Lam_list):
            delta_lam_k = Scalar(f'\\Delta\\lambda_{{{k}}}', codename=f'delta_lam_{k}', real=True)
            subs_dict[dot_lam] = delta_lam_k
        
        # Inequality constraint Lagrange multipliers
        for lam in lam_phi_list:
            lam_codename = self._get_codename(lam)
            delta_lam_phi = Scalar(f'\\Delta{{{lam.name}}}', codename=f'delta_{lam_codename}', real=True)
            subs_dict[lam] = delta_lam_phi
        
        # Time differential substitution
        if hasattr(self, 'd_t'):
            subs_dict[self.d_t] = delta_t
        else:
            # Create d_t symbol if not present
            d_t = Scalar(r'\mathrm{d}t', codename='dt', real=True)
            subs_dict[d_t] = delta_t
        
        return subs_dict

    @cached_property
    def delta_A_c_(self) -> sp.Matrix:
        """
        Augmented vector of unknowns containing increments of internal variables and Lagrange multipliers.
        
        Following gsm_engine.py pattern from _get_Sig_f_R_dR_n1 method:
        delta_A = Matrix([delta_Eps, Matrix(delta_Lam), Matrix(delta_lam)])
        
        This vector contains all the unknown increments that need to be solved
        in the nonlinear system of equations at each time step.
        
        The _c suffix indicates the augmented dimension (internal + constraints).
        
        Returns:
            Matrix containing all unknown increments
        """
        from bmcs_matmod.gsm_lagrange.core.gsm_vars import Scalar
        
        # Internal variable increments
        delta_Eps = sp.Matrix([
            Scalar(f'\\Delta{{{var.name}}}', codename=f'delta_{self._get_codename(var)}', real=True) 
            for var in self.Eps_b
        ])
        
        # Constraint multiplier increments
        lam_phi_list, lam_phi_term, _ = self.lam_phi_f_
        dot_Lam_list, delta_Lam, dot_lam_sum = self.H_Lam_
        
        # Equality constraint multiplier increments
        delta_Lam_matrix = sp.Matrix([
            Scalar(f'\\Delta\\lambda_{{{k}}}', codename=f'delta_lam_{k}', real=True)
            for k in range(len(self.h_k))
        ])
        
        # Inequality constraint multiplier increments
        delta_lam_matrix = sp.Matrix([
            Scalar(f'\\Delta{{{lam.name}}}', codename=f'delta_{self._get_codename(lam)}', real=True)
            for lam in lam_phi_list
        ])
        
        # Assemble complete augmented vector
        delta_A_components = [delta_Eps]
        if len(delta_Lam_matrix) > 0:
            delta_A_components.append(delta_Lam_matrix)
        if len(delta_lam_matrix) > 0:
            delta_A_components.append(delta_lam_matrix)
        
        if len(delta_A_components) == 1:
            return delta_A_components[0]
        else:
            return sp.Matrix.vstack(*delta_A_components)

    @cached_property
    def S_c_(self) -> sp.Matrix:
        """
        Augmented vector of generalized forces conjugate to delta_A_c_.
        
        Following gsm_engine.py pattern from _get_Sig_f_R_dR_n1 method:
        S = Matrix([Sig, Matrix(dot_Lam), Matrix(lam)])
        
        This vector contains all the generalized forces that are conjugate to
        the unknown increments in delta_A_c_.
        
        The _c suffix indicates the augmented dimension (internal + constraints).
        
        Returns:
            Matrix containing all generalized forces
        """
        # Internal thermodynamic forces
        Sig = self.Sig_b
        
        # Constraint multiplier components
        lam_phi_list, lam_phi_term, _ = self.lam_phi_f_
        dot_Lam_list, delta_Lam, dot_lam_sum = self.H_Lam_
        
        # Assemble complete augmented vector
        S_components = [Sig]
        if len(dot_Lam_list) > 0:
            S_components.append(sp.Matrix(dot_Lam_list))
        if len(lam_phi_list) > 0:
            S_components.append(sp.Matrix(lam_phi_list))
        
        if len(S_components) == 1:
            return S_components[0]
        else:
            return sp.Matrix.vstack(*S_components)

    @cached_property
    def R_c_n1_(self) -> sp.Matrix:
        """
        Residuum vector evaluated at time n+1 with time discretization substitutions.
        
        Following gsm_engine.py pattern from _get_Sig_f_R_dR_n1 method where
        the residuum is constructed from the derivative of the Lagrangian with
        respect to generalized forces, then time discretization substitutions
        are applied.
        
        This represents the nonlinear system of equations that must be solved
        to find the state at time n+1.
        
        The _c suffix indicates the augmented dimension (internal + constraints).
        
        Returns:
            Matrix representing the residuum that should equal zero
        """
        # Start with the continuous optimality criterion
        dL_dS_A = self.dL_dS_A_
        
        # Apply time discretization substitutions
        R_n1 = dL_dS_A.subs(self.subs_n1_)
        
        # Special handling for inequality constraint (flow potential)
        lam_phi_list, lam_phi_term, f_constraint = self.lam_phi_f_
        if lam_phi_list and self.f_expr != sp.S.Zero:
            # The last component should be the flow potential constraint f_
            # evaluated at time n+1
            f_n1 = self.f_expr.subs(self.subs_n1_)
            R_n1[-1] = f_n1
        
        return R_n1

    @cached_property
    def dR_dA_cc_n1_(self) -> sp.Matrix:
        """
        Jacobian matrix of the residuum with respect to the augmented unknowns at time n+1.
        
        Following gsm_engine.py pattern from _get_Sig_f_R_dR_n1 method where
        the Jacobian is computed as R_n1.jacobian(delta_A) with special handling
        for derivatives and Dirac delta functions.
        
        This matrix is used in Newton-Raphson iterations to solve the nonlinear
        system of equations at each time step.
        
        The _cc suffix indicates the expanded Jacobian dimensions (augmented × augmented).
        
        Returns:
            Jacobian matrix dR/dA for Newton-Raphson method
        """
        # Calculate Jacobian of residuum with respect to augmented unknowns
        dR_dA_n1 = self.R_c_n1_.jacobian(self.delta_A_c_)
        
        # Replace derivatives and Dirac delta functions with zeros
        # Following gsm_engine.py pattern
        dR_dA_n1 = dR_dA_n1.replace(sp.Derivative, lambda *args: 0)
        dR_dA_n1 = dR_dA_n1.replace(sp.DiracDelta, lambda *args: 0)
        
        return dR_dA_n1

    @cached_property
    def f_n1_(self) -> sp.Expr:
        """
        Flow potential (threshold function) evaluated at time n+1.
        
        Following gsm_engine.py pattern from _get_Sig_f_R_dR_n1 method where
        the static threshold function is substituted with constitutive relations
        and then time discretization substitutions are applied.
        
        This is used to determine if the material point is in the elastic or
        inelastic state (f ≤ 0 for admissible states).
        
        Returns:
            Flow potential expression at time n+1
        """
        if self.f_expr == sp.S.Zero:
            return sp.S.Zero
        
        # Apply constitutive relations first
        f_Eps = self.f_expr
        if self.subs_Sig_Eps_b:
            f_Eps = f_Eps.subs(self.subs_Sig_Eps_b)
        
        # Then apply time discretization substitutions
        f_n1 = f_Eps.subs(self.subs_n1_)
        
        return f_n1

    @cached_property
    def Sig_b_n1_(self) -> sp.Matrix:
        """
        Internal thermodynamic forces evaluated at time n+1.
        
        Following gsm_engine.py pattern from _get_Sig_f_R_dR_n1 method where
        Sig_n1 = Sig_.subs(subs_n1) represents the internal thermodynamic forces
        (not external stresses) evaluated with time discretization substitutions.
        
        This provides the internal force response at the end of the time step.
        
        The _b suffix indicates the internal variable dimension.
        
        Returns:
            Internal thermodynamic forces vector at time n+1
        """
        # Apply time discretization substitutions to internal thermodynamic force expressions
        Sig_n1 = self.Sig_b_.subs(self.subs_n1_)
        
        return Sig_n1

    @cached_property
    def sig_a_n1_(self) -> sp.Matrix:
        """
        External stress vector evaluated at time n+1.
        
        Following gsm_engine.py pattern where external stresses are evaluated
        with time discretization substitutions applied to the stress expressions
        derived from free energy derivatives.
        
        This provides the external stress response at the end of the time step.
        
        The _a suffix indicates the external variable dimension.
        
        Returns:
            External stress vector at time n+1
        """
        # Apply time discretization substitutions to external stress expressions
        sig_n1 = self.sig_a_.subs(self.subs_n1_)
        
        return sig_n1

    @cached_property
    def delta_eps_a_(self) -> sp.Matrix:
        """
        External strain increments as a vector (delta_eps_a_).
        
        Following the convention where _a index represents the expanded collection
        of all external strain increment variables as a homogeneous vector structure.
        This is the external counterpart to delta_A_c_ for internal variables.
        
        The trailing underscore indicates this is a derived expression used for
        time discretization and differentiation.
        
        Returns:
            Vector (Matrix) of external strain increment symbols
        """
        from bmcs_matmod.gsm_lagrange.core.gsm_vars import Scalar
        
        return sp.Matrix([
            Scalar(f'\\Delta{{{var.name}}}', codename=f'delta_{self._get_codename(var)}', real=True) 
            for var in self.eps_a
        ])

    @cached_property
    def dsig_deps_aa_n1_(self) -> sp.Matrix:
        """
        Material matrix: Jacobian of external stresses w.r.t. external strain increments at time n+1.
        
        Following gsm_engine.py pattern for algorithmic stiffness matrix construction where
        the material matrix is computed as sig_n1.jacobian(delta_eps) with special handling
        for derivatives and Dirac delta functions.
        
        This matrix represents the tangent stiffness relating external stress increments
        to external strain increments and is used in the global stiffness matrix assembly.
        
        The _aa suffix indicates the external × external dimensions (strain increments → stress increments).
        
        Returns:
            Material matrix dsig/deps for algorithmic stiffness
        """
        # Calculate Jacobian of external stresses with respect to external strain increments
        dsig_deps_n1 = self.sig_a_n1_.jacobian(self.delta_eps_a_)
        
        # Replace derivatives and Dirac delta functions with zeros
        # Following gsm_engine.py pattern
        dsig_deps_n1 = dsig_deps_n1.replace(sp.Derivative, lambda *args: 0)
        dsig_deps_n1 = dsig_deps_n1.replace(sp.DiracDelta, lambda *args: 0)
        
        return dsig_deps_n1

    def _get_codename(self, symbol: sp.Symbol) -> str:
        """
        Get codename for a symbol, with fallback to symbol name.
        
        Args:
            symbol: SymPy symbol to get codename for
        
        Returns:
            Codename string suitable for code generation
        """
        # First check if it's in any of our codename dictionaries
        for codename_dict in [
            self.eps_codenames, self.sig_codenames, 
            self.Eps_codenames, self.Sig_codenames, 
            self.param_codenames
        ]:
            if symbol in codename_dict:
                return codename_dict[symbol]
        
        # Check if symbol has codename attribute
        if hasattr(symbol, 'codename') and symbol.codename:
            return symbol.codename
            
        # Fall back to symbol name if it's a valid identifier
        if is_valid_variable_name(symbol.name):
            return symbol.name
            
        # Last resort: create a safe codename from symbol name
        safe_name = ''.join(c if c.isalnum() else '_' for c in symbol.name)
        return safe_name

    def latex_potentials(self) -> str:
        """
        Returns a LaTeX-friendly string with thermodynamic potentials and constitutive relations.
        
        Following the pattern from gsm_def.py but adapted for pure symbolic GSMSymbDef.
        Displays free energy, constitutive relations, and placeholders for future extensions.
        
        Returns:
            LaTeX string representation of the model potentials and relations
        """
        latex_lines = []
        
        # Model header
        latex_lines.append(f"## {self.__class__.__name__}")
        
        # Free energy
        latex_lines.append("### Free energy")
        if self.F_expr != sp.S.Zero:
            latex_lines.append(f"$$F = {sp.latex(self.F_expr)}$$")
        else:
            latex_lines.append("$$F = 0$$ (not defined)")
        
        # External constitutive relations
        if self.subs_sig_eps_a:
            latex_lines.append("### External constitutive relations")
            latex_lines.append("$$" + sp.latex(self.subs_sig_eps_a) + "$$")
        
        # Internal constitutive relations  
        if self.subs_Sig_Eps_b:
            latex_lines.append("### Internal constitutive relations")
            latex_lines.append("$$" + sp.latex(self.subs_Sig_Eps_b) + "$$")
        
        # Dissipation potential
        if self.phi_ != sp.S.Zero:
            latex_lines.append("### Dissipation potential")
            latex_lines.append(f"$$\\phi = {sp.latex(self.phi_)}$$")
        
        # Mechanical dissipation
        if self.gamma_mech_ != sp.S.Zero:
            latex_lines.append("#### Mechanical dissipation")
            latex_lines.append(f"$$\\gamma_{{\\mathrm{{mech}}}} = {sp.latex(self.gamma_mech_)}$$")
        else:
            latex_lines.append("#### Mechanical dissipation")
            latex_lines.append("$$\\gamma_{\\mathrm{mech}} = 0$$ (no internal variables)")
        
        # Lagrangian
        if self.L_ != sp.S.Zero:
            latex_lines.append("#### Lagrangian")  
            latex_lines.append(f"$$\\mathcal{{L}} = {sp.latex(self.L_)}$$")
        else:
            latex_lines.append("#### Lagrangian")  
            latex_lines.append("$$\\mathcal{L} = 0$$ (no constraints or dissipation)")
        
        # Optimality criterion
        if len(self.dL_dS_A_) > 0:
            latex_lines.append("#### Optimality criterion")
            latex_lines.append(f"$$\\frac{{\\partial \\mathcal{{L}}}}{{\\partial \\mathcal{{S}}}} = {sp.latex(self.dL_dS_A_)} = 0$$")
        else:
            latex_lines.append("#### Optimality criterion")
            latex_lines.append("$$\\frac{\\partial \\mathcal{L}}{\\partial \\mathcal{S}} = 0$$ (no generalized forces)")
        
        # Time discretization substitutions
        if self.subs_n1_:
            latex_lines.append("#### Time discretization substitutions")
            # Display only a subset to avoid overwhelming output
            subs_items = list(self.subs_n1_.items())[:6]  # Show first 6 substitutions
            subs_display = dict(subs_items)
            if len(subs_items) < len(self.subs_n1_):
                latex_lines.append(f"$$\\text{{(showing first 6 of {len(self.subs_n1_)} substitutions)}}$$")
            latex_lines.append("$$" + sp.latex(subs_display) + "$$")
        
        # Bounds of inelastic process (if applicable)
        if hasattr(self, 'dot_Eps_bounds_expr') and self.dot_Eps_bounds_expr != sp.S.Zero:
            latex_lines.append("#### Bounds of inelastic process")
            latex_lines.append("$$" + sp.latex(self.dot_Eps_bounds_expr) + " \\leq 0$$")
        
        return "\n".join(latex_lines)

    def print_potentials(self) -> None:
        """
        Print thermodynamic potentials in a formatted display.
        
        Convenience method for interactive use, displaying LaTeX expressions
        using IPython display capabilities when available.
        """
        print('=' * 50)
        print(f'class {self.__class__.__name__}')
        print('=' * 50)
        
        try:
            from IPython.display import display, Math
            print('Free energy:')
            display(Math(f'F = {sp.latex(self.F_expr)}'))
            
            if self.subs_sig_eps_a:
                print('External constitutive relations:')
                display(Math(sp.latex(self.subs_sig_eps_a)))
                
            if self.subs_Sig_Eps_b:
                print('Internal constitutive relations:')
                display(Math(sp.latex(self.subs_Sig_Eps_b)))
                
            if self.phi_ != sp.S.Zero:
                print('Dissipation potential:')
                display(Math(f'\\phi = {sp.latex(self.phi_)}'))
                
            if self.gamma_mech_ != sp.S.Zero:
                print('Mechanical dissipation:')
                display(Math(f'\\gamma_{{\\mathrm{{mech}}}} = {sp.latex(self.gamma_mech_)}'))
                
            if self.L_ != sp.S.Zero:
                print('Lagrangian:')
                display(Math(f'\\mathcal{{L}} = {sp.latex(self.L_)}'))
                
            if len(self.dL_dS_A_) > 0:
                print('Optimality criterion:')
                display(Math(f'\\frac{{\\partial \\mathcal{{L}}}}{{\\partial \\mathcal{{S}}}} = {sp.latex(self.dL_dS_A_)} = 0'))
                
            if self.subs_n1_:
                print('Time discretization substitutions:')
                # Display only first few to avoid overwhelming output
                subs_items = list(self.subs_n1_.items())[:4]
                subs_display = dict(subs_items)
                display(Math(sp.latex(subs_display)))
                if len(self.subs_n1_) > 4:
                    print(f'... (showing 4 of {len(self.subs_n1_)} total substitutions)')
                
        except ImportError:
            # Fallback for non-IPython environments
            print(f'Free energy: F = {self.F_expr}')
            if self.subs_sig_eps_a:
                print(f'External relations: {self.subs_sig_eps_a}')
            if self.subs_Sig_Eps_b:
                print(f'Internal relations: {self.subs_Sig_Eps_b}')
            if self.phi_ != sp.S.Zero:
                print(f'Dissipation: phi = {self.phi_}')
            if self.gamma_mech_ != sp.S.Zero:
                print(f'Mechanical dissipation: gamma_mech = {self.gamma_mech_}')
            if self.L_ != sp.S.Zero:
                print(f'Lagrangian: L = {self.L_}')
            if len(self.dL_dS_A_) > 0:
                print(f'Optimality criterion: dL_dS_A = {self.dL_dS_A_} = 0')
            if self.subs_n1_:
                print(f'Time discretization: {len(self.subs_n1_)} substitutions')



"""
GSM Widgets Package

This package provides interactive widgets for visualizing and manipulating
GSM (Generalized Standard Material) models using ipywidgets and ipytree.
"""

from .gsm_def_widget import GSMDefWidget, create_gsm_widget

__all__ = ['GSMDefWidget', 'create_gsm_widget']



from turtle import color
from urllib import response
from bmcs_matmod.api import GSMRM
import matplotlib.pylab as plt 
import sympy as sp
import numpy as np
from bmcs_utils.api import Cymbol, Model, mpl_align_yaxis_to_zero, Float, SymbExpr, InjectSymbExpr
# from bmcs_matmod.gsm.potentials.potential1d_t_e_vp_d import Potential1D_T_E_VP_D_SymbExpr
import traits.api as tr
from scipy.integrate import cumtrapz
from pathlib import Path
from scipy.interpolate import interp1d

class MatStudy_T_E_VP_D(Model):

    gsm_F = GSMRM.load_from_disk('gsm_F_1d_t_e_vp_d')
    gsm_G = GSMRM.load_from_disk('gsm_G_1d_t_e_vp_d')

    @staticmethod
    def gsm_run(gsm_, u_ta, T_t, t_t, **material_params):
        response = gsm_.get_response(u_ta, T_t, t_t, **material_params)
        _t_t, _u_tIa, _T_t, _Eps_tIb, _Sig_tIb, _iter_t, _dF_dEps_t, lam_t = response
        _u_atI, _Eps_btI, _Sig_btI, _dF_dEps_btI = [np.moveaxis(v_, -1, 0) for v_ in (_u_tIa, _Eps_tIb, _Sig_tIb, _dF_dEps_t)]
        _sig_atI = gsm_.get_sig(_u_atI, _T_t, _Eps_btI, _Sig_btI, **material_params )
        return _t_t, _u_atI, _sig_atI, _T_t, _Eps_btI, _Sig_btI, _dF_dEps_btI 

    gsm_F.vp_on = True
    gsm_F.update_at_k = False
    gsm_G.vp_on = True
    gsm_G.update_at_k = False

    material_params = tr.Dict
    def _material_params_default(self):
        _f_c = 44
        _f_t = -0.1 * _f_c
        _X_0 = (_f_c + _f_t) / 2
        _f_s = (_f_c - _f_t) / 2
        _E = 50000
        _KH_factor = 4
        _KH = _E * _KH_factor
        _K_ratio = 0.01 # 0.015
        _K = _KH * _K_ratio
        _H = _KH * (1 - _K_ratio)
        
        return dict(
            E=_E, 
            gamma_lin=_H, # _E * 10, 
            gamma_exp=0.5,
            alpha_0=0.5,
            K_lin=_K, # _E / 5,
            k_exp=10,
            z_0=10,
            S=0.008,
            c=2.5,
            r=2.7,
            f_c=_f_s,
            X_0=_X_0,  
            eta=500,
            T_0=20,
            C_v=0.01, # 0.0001, 
            beta=0.0001,
            alpha_therm=0, # 1.2e-5,
            d_N=1
        )

    dot_eps = Float(0.01, TIME=True)
    eps_max = Float(0.0035, TIME=True)

    monotonic_response = tr.Property(depends_on='state_changed')
    @tr.cached_property
    def _get_monotonic_response(self):
        # params
        print('recalculating')
        n_t = 151
        n_I = 1
        t_max = self.eps_max / self.dot_eps
        t_t_F = np.linspace(0, t_max, n_t)
        u_ta_F = (self.dot_eps * t_t_F).reshape(-1, 1)
        T_t = 20 + t_t_F * 0
        response_monotonic_F = self.gsm_run(self.gsm_F, u_ta_F, T_t, t_t_F, **self.material_params)
        _t_t_F, _u_atI_F, _sig_atI_F, _T_t_F, _Eps_btI_F, _Sig_btI_F, _dF_dEps_btI_F = response_monotonic_F
        _argmax_sig = np.argmax(_sig_atI_F)
        print('argmax_max_sig', _argmax_sig)
        print('shape', t_t_F.shape, _sig_atI_F.shape)
#        _max_sig = _sig_atI_F[_argmax_t]
        _max_sig = _sig_atI_F[0, _argmax_sig, 0]
        print('max_sig', _max_sig)
#        t_t_G = np.linspace(0, _t_t_F[_argmax_sig], n_t)
#        u_ta_G = (_max_sig / t_t).reshape(-1, 1)
        # u_ta_G = np.linspace(0, _max_sig, n_t).reshape(-1, 1)
        t_t_G = _t_t_F[:_argmax_sig+1]
        print('history', t_t_F[_argmax_sig], t_t_G[-1])
        u_ta_G = _sig_atI_F[0, :_argmax_sig+1, 0].reshape(-1, 1)
        T_t = 20 + t_t_G * 0
        # print('max_sig', _max_sig)
        # print('shape', len(t_t), t_t[-1], u_ta_G[-1])
        response_monotonic_G = self.gsm_run(self.gsm_G, u_ta_G, T_t, t_t_G, **self.material_params)
        _t_t_G, _u_atI_G, _sig_atI_G, _T_t_G, _Eps_btI_G, _Sig_btI_G, _dF_dEps_btI_G = response_monotonic_G
        print(_t_t_F[_argmax_sig], _t_t_G[-1])
        Diss_btI_F = cumtrapz(_dF_dEps_btI_F, _Eps_btI_F, initial=0, axis=1)
        Diss_btI_G = cumtrapz(_dF_dEps_btI_G, _Eps_btI_G, initial=0, axis=1)
        return response_monotonic_F, response_monotonic_G, (Diss_btI_F, Diss_btI_G, _max_sig)

    def get_fig_monotonic(self):
        fig, axes = plt.subplots(2,3, figsize=(12,6), tight_layout=True)
        return fig, axes

    def plot_monotonic(self, axes):
        (ax, ax_T, ax_Diss), (ax_omega, ax_3, ax_4) = axes
        response_F, response_G, (Diss_btI_F, Diss_btI_G, _max_sig) = self.monotonic_response
        _t_t_F, _u_atI_F, _sig_atI_F, _T_t_F, _Eps_btI_F, _Sig_btI_F, _dF_dEps_btI_F = response_F
        _t_t_G, _u_atI_G, _sig_atI_G, _T_t_G, _Eps_btI_G, _Sig_btI_G, _dF_dEps_btI_G = response_G
        _u_p_atI, _z_atI, _alpha_atI, _omega_atI = self.gsm_G.Eps_as_blocks(_Eps_btI_G)
        _, _Z_atI, _X_atI, _Y_atI = self.gsm_G.Eps_as_blocks(_Sig_btI_G)
        _arg_t_F = _t_t_F[np.argmax(_sig_atI_F)]
        _t_F_scale = _arg_t_F * _t_t_F[-1]

        ax_3_a = ax_3.twinx()
        ax_3.plot(_t_t_F, _u_atI_F[0, :, 0], label='Helmholtz', color='blue');
        ax_3_a.plot(_t_t_G, _u_atI_G[0, :, 0], label='Gibbs', color='orange');
        self._decorate_plot(ax_3, r'loading history', r'$t$')
        ax.plot(_u_atI_F[0, :, 0], _sig_atI_F[0, :, 0], label='Helmholtz', color='blue');
        ax.plot(_sig_atI_G[0, :, 0], _u_atI_G[0, :, 0], label='Gibbs', color='orange');
        self._decorate_plot(ax, r'stress-strain', r'$\varsigma$')
        ax_T.plot(_u_atI_F[0, :, 0], _T_t_F, label='Helmholtz', color='blue');
        ax_T.plot(_sig_atI_G[0, :, 0], _T_t_G, label='Gibbs', color='orange');
        self._decorate_plot(ax_T, r'temperature', r'$\vartheta$')
        ax_Diss.plot(_t_t_F, np.sum(Diss_btI_F[...,0], axis=0), alpha=1, label='F', color='blue')   
        ax_Diss.plot(_t_t_G * _t_F_scale, np.sum(Diss_btI_G[...,0], axis=0), alpha=1, label='G', color='orange')    

        ax_omega.plot(_sig_atI_G[0, :, 0], _omega_atI[0, :, 0], color='orange', lw=2, label='Gibbs')
        ax_omega.set_xlabel(r'$-\varepsilon$/-')
        ax_omega.set_ylabel(r'$\omega$/-')
        return

    # TODO Rename this here and in `plot_monotonic`
    def _decorate_plot(self, arg0, arg1, arg2):
        # ax_T.plot(_t_t_F, _T_t_F);
        # ax_T.plot(_t_t_G * _t_F_scale, _T_t_G);
        arg0.legend()
        arg0.set_title(arg1)
        arg0.set_ylabel(arg2)
        arg0.set_xlabel(r'$-\varepsilon$')

    @staticmethod
    def generate_cyclic_load(max_s, min_s, freq, total_cycles, points_per_cycle):
        # Calculate the time for one cycle
        total_time = total_cycles / freq

        # Calculate the mean value and amplitude
        mean_value = (max_s + min_s) / 2
        amplitude = (max_s - min_s) / 2

        # Calculate the initial loading slope
        slope = 2 * np.pi * freq * amplitude
        
        # Time arrays for linear increase and sinusoidal part
        initial_duration = mean_value / slope
        initial_points = int(initial_duration * freq * points_per_cycle)
        total_points = int(total_time * freq * points_per_cycle)
        
        # Generate the initial linear increase
        initial_t = np.linspace(0, initial_duration, initial_points, endpoint=False)
        initial_loading = slope * initial_t

        # Generate the sinusoidal loading
        sinusoidal_t = np.linspace(0, total_time, total_points, endpoint=False)
        sinusoidal_loading = mean_value + amplitude * np.sin(2 * np.pi * freq * sinusoidal_t)

        # Combine the initial linear increase with the sinusoidal loading
        t_full = np.concatenate((initial_t, sinusoidal_t + initial_duration))
        s_full = np.concatenate((initial_loading, sinusoidal_loading))
        
        return t_full, s_full

    @staticmethod    
    def arg_max_min(data):
        # Find local maxima and minima
        maxima_indexes = (np.diff(np.sign(np.diff(data))) < 0).nonzero()[0] + 1
        minima_indexes = (np.diff(np.sign(np.diff(data))) > 0).nonzero()[0] + 1
        return maxima_indexes, minima_indexes
    
    get_step_loading = tr.Property
    tr.cached_property
    def _get_get_step_loading(self):
        s_1 = sp.Symbol('s_1')
        t_1 = sp.Symbol('t_1')
        t = sp.Symbol('t')
        fn_s_t = sp.Piecewise((t * s_1/t_1, t < t_1),(s_1, True))
        return sp.lambdify((t, s_1, t_1), fn_s_t)

    def plot_loading(self):
        fig, (ax, ax_N) = plt.subplots(2,1, figsize=(8,6))
        t_t, s_t = self.generate_cyclic_load(1, 0.1, self.freq, 10, 30)
        ax.plot(t_t, s_t, '-o')
        arg_max, arg_min = self.arg_max_min(s_t)
        ax.plot(t_t[arg_max], s_t[arg_max], 'o', color='red')
        ax.plot(t_t[arg_min], s_t[arg_min], 'o', color='orange')
        ax_N.plot(s_t[arg_max], 'o-')
        ax_N.plot(s_t[arg_min], 'o-')

        u_t_fat = self.get_step_loading(t_t, 1, 50)
        ax.plot(t_t, u_t_fat)
        return
    
    get_sig_p_0 = tr.Property
    @tr.cached_property
    def _get_get_sig_p_0(self):
        sig_p = self.gsm_G.Sig_vars[0][0]
        sig_p_solved_ = sp.solve((self.gsm_G.f_), sig_p)
        return sp.lambdify((self.gsm_G.T_var, 
                            self.gsm_G.Eps.as_explicit(), 
                            self.gsm_G.Sig.as_explicit()) + self.gsm_G.m_params + ('**kw',),
                            sig_p_solved_, cse=True)

    S_max_levels = tr.Array(value = [1, 0.95, 0.85, 0.75, 0.65], dtype=np.float_, TIME=True)

    S_min = Float(0.1, TIME=True)

    freq = Float(5, TIME=True)

    total_cycles = Float(1000, TIME=True)

    points_per_cycle = Float(66, TIME=True)

    fatigue_responses = tr.Property(depends_on='state_changed')
    @tr.cached_property
    def _get_fatigue_responses(self):
        print(f'calculating fatigue for {self.freq}')
        response_F, response_G, (Diss_btI_F, _, _max_sig) = self.monotonic_response
        S_max_levels = self.S_max_levels
        responses = {}
        N_S_min = {}

        N_S_max = {1: 1}
        Diss_Sbt = {1: Diss_btI_F[:,:,0]}
        _delta_T = 20
        for S_max in S_max_levels[1:]:

            print('S_max', S_max)
            # params
            t_t, s_t = self.generate_cyclic_load(S_max, self.S_min, self.freq, self.total_cycles, self.points_per_cycle)
            u_ta_fat = (_max_sig * s_t).reshape(-1, 1)
            T_t = 20 + t_t * 0

            response = self.gsm_run(self.gsm_G, u_ta_fat, T_t, t_t, **self.material_params)
            responses[S_max] = response

            _t_t_fat, _u_atI_fat, _sig_atI_fat, _T_t_fat, _Eps_btI_fat, _Sig_btI_fat, _dF_dEps_btI_fat = response

            _sig_t_fat = _sig_atI_fat[0,:,0]
            arg_max_u, arg_min_u = self.arg_max_min(_sig_t_fat)
            _N_S_max, _N_S_min = len(arg_max_u), len(arg_min_u)
            Diss_btI_fat = cumtrapz(_dF_dEps_btI_fat, _Eps_btI_fat, initial=0, axis=1)
            N_S_max[S_max] = _N_S_max
            N_S_min[S_max] = _N_S_min
            Diss_Sbt[S_max] = Diss_btI_fat[:, :, 0]

        _Diss_plastic_St = {S_max : np.sum(Diss_Sbt[S_max][:-1, :], axis=0) for S_max in S_max_levels}
        _Diss_damage_St = {S_max : Diss_Sbt[S_max][-1, :] for S_max in S_max_levels}

        _Diss_plastic_S = np.array([_Diss_plastic_St[S_max][-1] for S_max in S_max_levels])
        _Diss_damage_S = np.array([_Diss_damage_St[S_max][-1] for S_max in S_max_levels])
        return responses, N_S_max, N_S_min, _Diss_plastic_S, _Diss_damage_S

    def get_fig_SN(self):
        if self.include_last_col:
            fig, ((ax_sig_u, ax_el, ax_T), (ax_omega, ax_u, ax_SN)) = plt.subplots(2,3, figsize=(12,7), tight_layout=True)
        else:
            fig, ((ax_sig_u, ax_el), (ax_omega, ax_u)) = plt.subplots(2,2, figsize=(12,7), tight_layout=True)

        fig.canvas.header_visible=False

        ax_sig_u.set_title('stress-strain')
        ax_omega.set_title('damage')
        ax_el.set_title('elastic domain')
        ax_u.set_title('fatigue strain')

        ax_sig_u.set_xlabel(r'$-\varepsilon$/-')
        ax_sig_u.set_ylabel(r'$-\sigma$/MPa')
        ax_omega.set_xlabel(r'$\eta$/-')
        ax_omega.set_ylabel(r'$\omega$/-')
        ax_el.set_xlabel(r'$\eta$/-')
        ax_el.set_ylabel(r'$-\sigma$/MPa')
        ax_u.set_xlabel(r'$\eta$/-')
        ax_u.set_ylabel(r'$-\varepsilon$/-')

        if self.include_last_col:
            ax_T.set_title(r'temperature')
            ax_SN.set_title(r'S-N \& dissipation')

            ax_T.set_xlabel(r'$\eta$/-')
            ax_T.set_ylabel(r'$\vartheta$/$^{\circ}$C')

            ax_SN.set_xlabel(r'log$N$/-')
            ax_SN.set_ylabel(r'$S_\mathrm{max}$/-')

            ax_SN_twin = ax_SN.twinx()
            ax_SN_twin.set_ylabel(r'$\mathcal{D}/$J/mm$^{{2}}$')

            return fig, (ax_sig_u, ax_el, ax_T, ax_u, ax_omega, ax_SN, ax_SN_twin) 
        else:
            return fig, (ax_sig_u, ax_el, None, ax_u, ax_omega, None, None) 

    include_last_col = tr.Bool(True)

    def plot_fatigue_load(self, axes, S_max, c, alpha_line):
        responses, N_S_max, N_S_min, _Diss_plastic_S, _Diss_damage_S = self.fatigue_responses
        response = responses[S_max]
        _t_t_fat, _u_atI_fat, _sig_atI_fat, _T_t_fat, _Eps_btI_fat, _Sig_btI_fat, _dF_dEps_btI_fat = response
        ax_load, = axes
        ax_load.plot(_t_t_fat, _u_atI_fat[0,:,0], color=c, alpha=alpha_line, label=f'$S_{{\max}} = {S_max}$' )

    def plot_fatigue_response(self, axes, S_max, c, alpha_line):
        # params

        responses, N_S_max, N_S_min, _Diss_plastic_S, _Diss_damage_S = self.fatigue_responses
        response = responses[S_max]

        ax_sig_u, ax_el, ax_T, ax_u, ax_omega, ax_SN, ax_SN_twin = axes

        _t_t_fat, _u_atI_fat, _sig_atI_fat, _T_t_fat, _Eps_btI_fat, _Sig_btI_fat, _dF_dEps_btI_fat = response
        _sig_atI_top, _sig_atI_bot = self.get_sig_p_0(_T_t_fat, _Eps_btI_fat, _Sig_btI_fat, **self.material_params )
        _u_p_atI, _z_atI, _alpha_atI, _omega_atI = self.gsm_G.Eps_as_blocks(_Eps_btI_fat)
        _, _Z_atI, _X_atI, _Y_atI = self.gsm_G.Eps_as_blocks(_Sig_btI_fat)

        _sig_t_fat = _sig_atI_fat[0,:,0]
        _eta_t_fat = _t_t_fat / _t_t_fat[-1]

        _sig_t_fat = _sig_atI_fat[0,:,0]
        arg_max_u, arg_min_u = self.arg_max_min(_sig_t_fat)

        _sig_t_fat = _sig_atI_fat[0,:,0]
        _u_t_fat = _u_atI_fat[0,:,0]
        ax_sig_u.plot(_sig_t_fat, _u_t_fat, color=c, alpha=alpha_line, label=f'$S_{{\max}} = {S_max}$' )

        ax_omega.plot(_eta_t_fat, _omega_atI[0, :, 0], color=c, lw=2, label=f'$N = {N_S_max[S_max]}$' )

        ax_el.plot(_eta_t_fat, _sig_atI_top[:, 0], color=c, alpha=alpha_line)
        ax_el.plot(_eta_t_fat, _sig_atI_bot[:, 0], color=c, alpha=alpha_line)
        ax_el.fill_between(_eta_t_fat, _sig_atI_bot[:, 0], _sig_atI_top[:, 0], color=c, alpha=0.05)

        _eta_max_n = np.linspace(0, 1, N_S_max[S_max])
        _eta_min_n = np.linspace(0, 1, N_S_min[S_max])
        ax_u.plot(_eta_max_n, _sig_t_fat[arg_max_u], '-', lw=2, color=c)
        ax_u.plot(_eta_min_n, _sig_t_fat[arg_min_u], '--', lw=2, color=c)
        # if N_S_min[S_max] == N_S_max[S_max]:
        #     ax_el.fill_between(_eta_max_n, _sig_t_fat[arg_min_u], _sig_t_fat[arg_max_u], color=c, alpha=0.1)
        # else:
        if len(_eta_min_n) > 0 and len(_eta_max_n) > 0:
            f_min = interp1d(_eta_min_n, _sig_t_fat[arg_min_u], kind='linear', fill_value="extrapolate")
            f_max = interp1d(_eta_max_n, _sig_t_fat[arg_max_u], kind='linear', fill_value="extrapolate")
            _eta_common = np.linspace(0, 1, max(N_S_max[S_max], N_S_min[S_max]))
            ax_u.fill_between(_eta_common, f_min(_eta_common), f_max(_eta_common), color=c, alpha=0.05)

        if ax_T:
            ax_T.plot(_eta_t_fat, _T_t_fat, lw=2, color=c)

        arg_dissip = np.where( _dF_dEps_btI_fat[-1,:,0] > 1e-4 )
        # ax_sig_u.plot(_sig_t_fat[arg_dissip], _u_t_fat[arg_dissip], 'o', color='yellow', markersize=2)
        # ax_el.plot(_eta_t_fat[arg_dissip], _u_t_fat[arg_dissip], 'o', color='orange', markersize=3)
        # ax_el.plot(_eta_t_fat, _u_t_fat, 'o', color='yellow', markersize=2)

    def plot_S_max_study(self, S_max_select, include_last_col):
        # def plot_S_max_study(S_max_levels, responses, N_S_max, S_max_select=None, include_last_col=True):
        if S_max_select is None:
            S_max_select = np.arange(len(self.S_max_levels))

        fig_ax = self.get_fig_SN()
        fig, axes = fig_ax
        ax_sig_u, ax_el, ax_T, ax_u, ax_omega, ax_SN, ax_SN_twin = axes
        alpha_line = 0.5

        response_F, response_G, _ = self.monotonic_response
        _t_t_F, _u_atI_F, _sig_atI_F, _T_t_F, _Eps_btI_F, _Sig_btI_F, _dF_dEps_btI_F = response_F
        _t_t_G, _u_atI_G, _sig_atI_G, _T_t_G, _Eps_btI_G, _Sig_btI_G, _dF_dEps_btI_G = response_G

        ax_sig_u.plot(_u_atI_F[0,:,0], _sig_atI_F[0,:,0], color='black', ls='dashed')
        ax_sig_u.plot(_sig_atI_G[0,:,0], _u_atI_G[0,:,0], color='black')

        colors = ['black', 'red', 'darkslategrey', 'maroon', 'darkblue', 'magenta']

        for i in S_max_select:
            S_max = self.S_max_levels[i+1]
            c = colors[i+1]
            self.plot_fatigue_response(axes, S_max, c, alpha_line)

        if ax_SN:
            responses, N_S_max, N_S_min, _Diss_plastic_S, _Diss_damage_S = self.fatigue_responses
            self.plot_SN_curve(ax_SN, N_S_max, self.S_max_levels, ax_SN_twin,
                                                     _Diss_plastic_S, _Diss_damage_S)
        ax_u.set_ylim(ymin=0)

        ax_sig_u.legend()
        ax_omega.legend()
        #ax_el.set_ylim(ymax=_max_sig)
        return fig

    # TODO Rename this here and in `plot_S_max_study`
    def plot_SN_curve(self, ax_SN, N_S_max, S_max_levels, ax_SN_twin,
                                            _Diss_plastic_S, _Diss_damage_S):
        _N_S_max = np.array([N_S_max[key] for key in S_max_levels])
        ax_SN.semilogx(_N_S_max, S_max_levels, 'o-', lw=2, label='S-N')
        ax_SN.legend()
        ax_SN_twin.semilogx(_N_S_max, _Diss_plastic_S, 'o-', alpha=0.4, color='red', label='plastic')
        ax_SN_twin.semilogx(_N_S_max, _Diss_damage_S, 'o-', alpha=0.4, color='gray', label='damage')
        ax_SN_twin.legend()
        ax_SN_twin.set_xscale('log')

    path = Path().home() / 'simdb' / 'data'
    plot_config = {
    #     'one': ([0], True),
        # 'two': ([1,3], False),
        # 'endurance' : ([1,4], False), 
        # 'three': ([1,2, 3], False),
        # 'four': ([0, 1,2, 3], False),
        'all': ([0, 1, 2, 3], True),
    }

    def plot_fatigue_config(self):
        for name, config in self.plot_config.items():
            S_max_select, include_last_col = config
            fig = self.plot_S_max_study(S_max_select=S_max_select, include_last_col=include_last_col);
            fname = f'GSM_demo_fatigue_uniaxial_stress_{name}.png'
            fig.savefig(self.path / fname)
            fname = f'GSM_demo_fatigue_uniaxial_stress_{name}.pdf'
            fig.savefig(self.path / fname)
        return


#!/usr/bin/env python3
"""
AiiDA Integration Architecture - Implementation Plan

Based on analysis in notebooks/aiida_integration_strategies.ipynb, here's the 
recommended refactoring to properly handle AiiDA profiles and entry points.

CURRENT ISSUES:
1. ResponseDataNode in core/ tries to create AiiDA nodes without profile checks
2. CLIResponseData redundancy (already removed)
3. Missing entry point definitions for direct Python integration
4. No graceful degradation when AiiDA unavailable

RECOMMENDED CHANGES:
"""

# 1. MOVE ResponseDataNode to aiida_plugin/
"""
BEFORE:
bmcs_matmod/gsm_lagrange/core/response_data_node.py

AFTER:  
bmcs_matmod/gsm_lagrange/aiida_plugin/response_data_node.py

RATIONALE:
- Separates AiiDA-dependent code from core functionality
- Makes it clear that this module requires AiiDA
- Allows core/ to remain AiiDA-free
"""

# 2. ADD entry points to setup.cfg
setup_cfg_additions = """
[options.entry_points]
# Console scripts for CLI access
console_scripts =
    gsm-simulate = bmcs_matmod.gsm_lagrange.cli.cli_gsm:main

# AiiDA calculation entry points (direct Python integration)
aiida.calculations =
    gsm.lagrange = bmcs_matmod.gsm_lagrange.aiida_plugin.calculations:GSMLagrangeCalculation

# AiiDA data entry points
aiida.data =
    gsm.response = bmcs_matmod.gsm_lagrange.aiida_plugin.response_data_node:ResponseDataNode

# AiiDA workflow entry points  
aiida.workflows =
    gsm.material_characterization = bmcs_matmod.gsm_lagrange.aiida_plugin.workflows:MaterialCharacterizationWorkChain

# Custom plugin discovery (enables: pkg_resources.iter_entry_points('bmcs_matmod.gsm_models'))
bmcs_matmod.gsm_models =
    gsm1d_ed = bmcs_matmod.gsm_lagrange.models.gsm1d_ed:GSM1D_ED
    gsm1d_ep = bmcs_matmod.gsm_lagrange.models.gsm1d_ep:GSM1D_EP
    gsm1d_epd = bmcs_matmod.gsm_lagrange.models.gsm1d_epd:GSM1D_EPD
    gsm1d_evp = bmcs_matmod.gsm_lagrange.models.gsm1d_evp:GSM1D_EVP
    gsm1d_evpd = bmcs_matmod.gsm_lagrange.models.gsm1d_evpd:GSM1D_EVPD
    gsm1d_ve = bmcs_matmod.gsm_lagrange.models.gsm1d_ve:GSM1D_VE
    gsm1d_ved = bmcs_matmod.gsm_lagrange.models.gsm1d_ved:GSM1D_VED
    gsm1d_vevp = bmcs_matmod.gsm_lagrange.models.gsm1d_vevp:GSM1D_VEVP
    gsm1d_vevpd = bmcs_matmod.gsm_lagrange.models.gsm1d_vevpd:GSM1D_VEVPD
"""

# 3. CREATE aiida_plugin/ directory structure
aiida_plugin_structure = """
bmcs_matmod/gsm_lagrange/aiida_plugin/
├── __init__.py                 # Safe imports with fallbacks
├── response_data_node.py       # Moved from core/
├── calculations.py             # AiiDA Calculation classes
├── workflows.py                # AiiDA WorkChain classes
├── utils.py                    # Profile management utilities
└── parsers.py                  # Output parsers for calculations
"""

# 4. UPDATE CLI to use aiida_plugin
cli_updates = """
# OLD: from ..core.response_data_node import ResponseDataNode
# NEW: 
try:
    from ..aiida_plugin.response_data_node import ResponseDataNode, create_response_data_node
    RESPONSE_DATA_NODE_AVAILABLE = True
except ImportError:
    ResponseDataNode = None
    create_response_data_node = None
    RESPONSE_DATA_NODE_AVAILABLE = False
"""

# 5. BENEFITS of this architecture
benefits = """
✓ CLEAN SEPARATION: Core modules have zero AiiDA dependencies
✓ GRACEFUL DEGRADATION: Package works without AiiDA installed
✓ MULTIPLE INTEGRATION PATHS: CLI + direct Python entry points
✓ DISCOVERABILITY: Entry points enable plugin discovery
✓ PROFILE MANAGEMENT: Automatic profile loading with fallbacks
✓ MAINTAINABILITY: Clear module boundaries and responsibilities
"""

# 6. USAGE EXAMPLES after refactoring

# Standalone usage (no AiiDA required)
standalone_example = """
from bmcs_matmod.gsm_lagrange.core.gsm_model import GSMModel
from bmcs_matmod.gsm_lagrange.models.gsm1d_ed import GSM1D_ED

material = GSMModel(GSM1D_ED)
response_data = material.get_F_response(strain_history, time_array)
# Works without AiiDA!
"""

# AiiDA integration (optional)
aiida_example = """
from bmcs_matmod.gsm_lagrange.core.gsm_model import GSMModel
from bmcs_matmod.gsm_lagrange.models.gsm1d_ed import GSM1D_ED

# Core simulation (no AiiDA)
material = GSMModel(GSM1D_ED)
response_data = material.get_F_response(strain_history, time_array)

# Optional AiiDA storage
try:
    from bmcs_matmod.gsm_lagrange.aiida_plugin.response_data_node import create_response_data_node
    node = create_response_data_node(response_data, auto_load_profile=True)
    print(f"Stored in AiiDA: {node.uuid}")
except ImportError:
    print("AiiDA not available, using JSON storage")
    with open("results.json", "w") as f:
        json.dump(response_data.to_dict(), f)
"""

# Entry point usage (no CLI needed!)
entry_point_example = """
import pkg_resources

# Discover GSM models via entry points
for entry_point in pkg_resources.iter_entry_points('bmcs_matmod.gsm_models'):
    model_class = entry_point.load()
    print(f"Found model: {entry_point.name} -> {model_class}")

# Discover AiiDA calculations via entry points
for entry_point in pkg_resources.iter_entry_points('aiida.calculations'):
    if entry_point.name.startswith('gsm.'):
        calc_class = entry_point.load()
        print(f"Found AiiDA calculation: {entry_point.name} -> {calc_class}")
"""

if __name__ == "__main__":
    print("=== AiiDA Integration Architecture Plan ===")
    print("\nSee notebooks/aiida_integration_strategies.ipynb for full analysis.")
    print("\nKey recommendations:")
    print("1. Move ResponseDataNode to aiida_plugin/")
    print("2. Add entry points for discoverability")
    print("3. Implement graceful AiiDA profile handling")
    print("4. Keep core/ modules AiiDA-free")
    print("5. Use entry points for direct Python integration")
    print("\nThis enables both standalone and AiiDA-integrated usage!")



# from .ms1 import MS1
# from .ms1 import MS13D
# from .msx import MSX
# from .gsm import GSMRM
# from .gsm import GSMNR

from .gsm_lagrange.core.gsm_engine import GSMEngine
from .gsm_lagrange.core.gsm_def import GSMDef
from .gsm_lagrange.core.gsm_model import GSMModel
from .gsm_lagrange.models.gsm1d_ed import GSM1D_ED
from .gsm_lagrange.models.gsm1d_ep import GSM1D_EP
from .gsm_lagrange.models.gsm1d_epd import GSM1D_EPD
from .gsm_lagrange.models.gsm1d_evp import GSM1D_EVP
from .gsm_lagrange.models.gsm1d_evpd import GSM1D_EVPD
from .gsm_lagrange.models.gsm1d_ve import GSM1D_VE
from .gsm_lagrange.models.gsm1d_ved import GSM1D_VED
from .gsm_lagrange.models.gsm1d_vevp import GSM1D_VEVP
from .gsm_lagrange.models.gsm1d_vevpd import GSM1D_VEVPD

from .gsm_lagrange.core.response_data import ResponseData

from .time_function.time_fn import TimeFnBase, TimeFnStepLoading, TimeFnMonotonicAscending, \
    TimeFnCyclicBase, TimeFnCycleSinus, TimeFnCycleLinear, TimeFnCycleWithRamps, \
    TimeFnPeriodic, TimeFnStepping, TimeFnOverlay



#!/usr/bin/env python3
"""
AiiDA DataNode for GSM Response Data

This module provides a persistent, provenance-tracked storage solution for GSM simulation results.
It extends AiiDA's ArrayData to store ResponseData with metadata and provides a shared interface
for visualization and analysis.
"""

from typing import Dict, Any, Optional, TYPE_CHECKING
import numpy as np
from ..core.response_data_viz import ResponseDataVisualizationMixin

if TYPE_CHECKING:
    from ..core.response_data import ResponseData

try:
    from aiida.orm import ArrayData
    from aiida.common import AttributeDict
    from aiida.manage import get_manager
    import aiida
    AIIDA_AVAILABLE = True
    
    # Check if profile is loaded on import
    try:
        get_manager().get_profile()
        AIIDA_PROFILE_READY = True
    except Exception:
        AIIDA_PROFILE_READY = False
        
except ImportError:
    # Fallback base class when AiiDA is not available
    class ArrayData:
        def __init__(self):
            raise ImportError("AiiDA not available. Install aiida-core to use ResponseDataNode.")
    AIIDA_AVAILABLE = False
    AIIDA_PROFILE_READY = False
    aiida = None


class ResponseDataNode(ArrayData, ResponseDataVisualizationMixin):
    """
    AiiDA DataNode for persistent storage of GSM simulation response data.
    
    This class provides:
    - Immutable, versioned storage of simulation results
    - Automatic provenance tracking
    - Database queryability
    - Shared interface with ResponseData for visualization
    - Efficient binary storage of large arrays
    
    The class maintains the same interface as ResponseData for seamless
    integration with plotting and analysis tools.
    """
    
    def __init__(self, **kwargs):
        """Initialize ResponseDataNode with profile checking"""
        if not AIIDA_AVAILABLE:
            raise ImportError("AiiDA not available. Install aiida-core to use ResponseDataNode.")
        
        # Check if profile is loaded, try to load if not
        if not AIIDA_PROFILE_READY:
            try:
                aiida.load_profile()
            except Exception as e:
                raise RuntimeError(
                    f"AiiDA profile not loaded and auto-load failed: {e}\n"
                    "Please load a profile with aiida.load_profile() before creating ResponseDataNode."
                )
        
        super().__init__(**kwargs)
    
    @classmethod
    def from_response_data(cls, response_data: 'ResponseData', 
                          simulation_metadata: Optional[Dict[str, Any]] = None) -> 'ResponseDataNode':
        """
        Create a ResponseDataNode from active ResponseData.
        
        Parameters
        ----------
        response_data : ResponseData
            The active simulation data to store
        simulation_metadata : dict, optional
            Additional metadata about the simulation (parameters, model info, etc.)
            
        Returns
        -------
        ResponseDataNode
            Persistent storage node with all simulation data and metadata
        """
        node = cls()
        
        # Store main time series arrays
        node.set_array('time', response_data.t_t)
        node.set_array('eps', response_data.eps_t)
        node.set_array('sig', response_data.sig_t)
        node.set_array('iterations', response_data.iter_t)
        node.set_array('lambda', response_data.lam_t)
        
        # Store flattened internal variables and forces
        node.set_array('Eps_t_flat', response_data.Eps_t_flat)
        node.set_array('Sig_t_flat', response_data.Sig_t_flat)
        
        # Store individual variables with codenames
        for var_name, var_data in response_data.Eps_t.items():
            if isinstance(var_data, np.ndarray):
                node.set_array(f'Eps_{var_name}', var_data)
        
        for var_name, var_data in response_data.Sig_t.items():
            if isinstance(var_data, np.ndarray):
                node.set_array(f'Sig_{var_name}', var_data)
        
        # Store metadata as queryable attributes
        node.base.attributes.set('simulation_info', {
            'n_steps': len(response_data.t_t),
            'time_range': [float(response_data.t_t[0]), float(response_data.t_t[-1])],
            'eps_range': [float(np.min(response_data.eps_t)), float(np.max(response_data.eps_t))],
            'sig_range': [float(np.min(response_data.sig_t)), float(np.max(response_data.sig_t))]
        })
        
        # Store variable metadata
        node.base.attributes.set('variable_info', {
            'Eps_vars': [getattr(var, 'codename', str(var)) for var in response_data.Eps_vars],
            'Sig_vars': [getattr(var, 'codename', str(var)) for var in response_data.Sig_vars],
            'Eps_codenames': response_data.Eps_codenames,
            'Sig_codenames': response_data.Sig_codenames,
            'Eps_var_shapes': [getattr(var, 'shape', ()) for var in response_data.Eps_vars],
            'Sig_var_shapes': [getattr(var, 'shape', ()) for var in response_data.Sig_vars],
            'Eps_var_names': [str(var) for var in response_data.Eps_vars],
            'Sig_var_names': [str(var) for var in response_data.Sig_vars]
        })
        
        # Store simulation metadata if provided
        if simulation_metadata:
            node.base.attributes.set('simulation_metadata', simulation_metadata)
        
        return node
    
    def to_response_data(self) -> 'ResponseData':
        """
        Convert stored data back to active ResponseData format.
        
        Returns
        -------
        ResponseData
            Active simulation data reconstructed from storage
        """
        from ..core.response_data import ResponseData, ResponseDataContainer
        
        # Get main arrays
        t_t = self.get_array('time')
        eps_t = self.get_array('eps')
        sig_t = self.get_array('sig')
        iter_t = self.get_array('iterations')
        lam_t = self.get_array('lambda')
        Eps_t_flat = self.get_array('Eps_t_flat')
        Sig_t_flat = self.get_array('Sig_t_flat')
        
        # Reconstruct variable metadata (simplified - would need proper Variable objects)
        var_info = self.base.attributes.get('variable_info')
        
        # Create mock variable objects with essential attributes
        class MockVariable:
            def __init__(self, codename, shape, name):
                self.codename = codename
                self.shape = shape
                self._name = name
            def __str__(self):
                return self._name
        
        Eps_vars = tuple(MockVariable(name, tuple(shape), var_name) 
                        for name, shape, var_name in zip(
                            var_info['Eps_vars'],
                            var_info['Eps_var_shapes'], 
                            var_info['Eps_var_names']))
        
        Sig_vars = tuple(MockVariable(name, tuple(shape), var_name)
                        for name, shape, var_name in zip(
                            var_info['Sig_vars'],
                            var_info['Sig_var_shapes'],
                            var_info['Sig_var_names']))
        
        # Reconstruct individual variable dictionaries
        Eps_t_dict = {}
        for var_name in var_info['Eps_vars']:
            try:
                Eps_t_dict[var_name] = self.get_array(f'Eps_{var_name}')
            except KeyError:
                # Fallback to reconstruction from flat arrays if individual arrays not found
                pass
        
        Sig_t_dict = {}
        for var_name in var_info['Sig_vars']:
            try:
                Sig_t_dict[var_name] = self.get_array(f'Sig_{var_name}')
            except KeyError:
                # Fallback to reconstruction from flat arrays if individual arrays not found
                pass
        
        # Create ResponseData instance
        return ResponseData(
            t_t=t_t,
            eps_t=eps_t,
            sig_t=sig_t,
            Eps_t_flat=Eps_t_flat,
            Sig_t_flat=Sig_t_flat,
            Eps_vars=Eps_vars,
            Sig_vars=Sig_vars,
            Eps_codenames=var_info['Eps_codenames'],
            Sig_codenames=var_info['Sig_codenames'],
            Eps_t=ResponseDataContainer(Eps_t_dict),
            Sig_t=ResponseDataContainer(Sig_t_dict),
            iter_t=iter_t,
            lam_t=lam_t
        )
    
    # Shared interface methods for visualization compatibility
    @property
    def t_t(self):
        """Time array"""
        return self.get_array('time')
    
    @property
    def eps_t(self):
        """Strain array"""
        return self.get_array('eps')
    
    @property
    def sig_t(self):
        """Stress array"""
        return self.get_array('sig')
    
    @property
    def Eps_t(self):
        """Internal variables dictionary-like access"""
        var_info = self.base.attributes.get('variable_info')
        result = {}
        for var_name in var_info['Eps_vars']:
            try:
                result[var_name] = self.get_array(f'Eps_{var_name}')
            except KeyError:
                pass
        return result
    
    @property
    def Sig_t(self):
        """Thermodynamic forces dictionary-like access"""
        var_info = self.base.attributes.get('variable_info')
        result = {}
        for var_name in var_info['Sig_vars']:
            try:
                result[var_name] = self.get_array(f'Sig_{var_name}')
            except KeyError:
                pass
        return result
    
    def get_Eps_array(self, key: str):
        """Get internal variable by codename (compatible with ResponseData)"""
        return self.get_array(f'Eps_{key}')
    
    def get_Sig_array(self, key: str):
        """Get conjugate variable by codename (compatible with ResponseData)"""
        return self.get_array(f'Sig_{key}')
    
    def get_simulation_info(self) -> Dict[str, Any]:
        """Get simulation summary information"""
        return self.base.attributes.get('simulation_info', {})
    
    def get_variable_info(self) -> Dict[str, Any]:
        """Get variable metadata information"""
        return self.base.attributes.get('variable_info', {})
    
    def get_simulation_metadata(self) -> Dict[str, Any]:
        """Get additional simulation metadata"""
        return self.base.attributes.get('simulation_metadata', {})
    
    def __repr__(self) -> str:
        """String representation"""
        try:
            info = self.get_simulation_info()
            var_info = self.get_variable_info()
            return (
                f"ResponseDataNode(uuid={self.uuid}, "
                f"n_steps={info.get('n_steps', 'unknown')}, "
                f"Eps_vars={var_info.get('Eps_vars', [])}, "
                f"Sig_vars={var_info.get('Sig_vars', [])})"
            )
        except:
            return f"ResponseDataNode(uuid={self.uuid})"
    
    def to_json_dict(self, include_arrays=True, include_metadata=True):
        """
        Convert ResponseDataNode to a JSON-serializable dictionary.
        
        Parameters
        ----------
        include_arrays : bool
            Whether to include full time series arrays
        include_metadata : bool
            Whether to include variable metadata
            
        Returns
        -------
        dict
            JSON-serializable dictionary with simulation results
        """
        result = {
            "status": "success",
            "simulation_info": {
                "n_steps": len(self.t_t),
                "time_range": [float(self.t_t[0]), float(self.t_t[-1])],
                "eps_range": [float(np.min(self.eps_t)), float(np.max(self.eps_t))],
                "sig_range": [float(np.min(self.sig_t)), float(np.max(self.sig_t))]
            }
        }
        
        if include_arrays:
            result["time_series"] = {
                "time": self.t_t.tolist(),
                "strain": self.eps_t[:, 0].tolist() if self.eps_t.ndim > 1 else self.eps_t.tolist(),
                "stress": self.sig_t[:, 0, 0].tolist() if self.sig_t.ndim > 2 else self.sig_t.tolist()
            }
            
            # Add internal variables using codenames
            result["internal_variables"] = {}
            for var_name, var_data in self.Eps_t.items():
                if isinstance(var_data, np.ndarray):
                    if var_data.ndim == 1:
                        result["internal_variables"][var_name] = var_data.tolist()
                    elif var_data.ndim == 2 and var_data.shape[1] == 1:
                        result["internal_variables"][var_name] = var_data[:, 0].tolist()
                    else:
                        # Flatten for higher dimensions
                        flattened = var_data.reshape(var_data.shape[0], -1)
                        if flattened.shape[1] == 1:
                            result["internal_variables"][var_name] = flattened[:, 0].tolist()
                        else:
                            result["internal_variables"][var_name] = flattened.tolist()
            
            # Add thermodynamic forces using codenames
            result["thermodynamic_forces"] = {}
            for var_name, var_data in self.Sig_t.items():
                if isinstance(var_data, np.ndarray):
                    if var_data.ndim == 1:
                        result["thermodynamic_forces"][var_name] = var_data.tolist()
                    elif var_data.ndim == 2 and var_data.shape[1] == 1:
                        result["thermodynamic_forces"][var_name] = var_data[:, 0].tolist()
                    else:
                        # Flatten for higher dimensions
                        flattened = var_data.reshape(var_data.shape[0], -1)
                        if flattened.shape[1] == 1:
                            result["thermodynamic_forces"][var_name] = flattened[:, 0].tolist()
                        else:
                            result["thermodynamic_forces"][var_name] = flattened.tolist()
        
        if include_metadata:
            var_info = self.get_variable_info()
            result["metadata"] = {
                "eps_variables": var_info.get('Eps_vars', []),
                "sig_variables": var_info.get('Sig_vars', []),
                "eps_var_shapes": var_info.get('Eps_var_shapes', []),
                "sig_var_shapes": var_info.get('Sig_var_shapes', [])
            }
        
        # Add AiiDA-specific metadata
        result["aiida_metadata"] = {
            "uuid": str(self.uuid),
            "node_type": "ResponseDataNode",
            "is_stored": self.is_stored,
            "creation_time": self.ctime.isoformat() if hasattr(self, 'ctime') and self.ctime else None
        }
        
        return result
    
    def to_json_string(self, include_arrays=True, include_metadata=True, indent=2):
        """
        Convert ResponseDataNode to a JSON string.
        
        Parameters
        ----------
        include_arrays : bool
            Whether to include full time series arrays
        include_metadata : bool
            Whether to include variable metadata
        indent : int
            JSON indentation level
            
        Returns
        -------
        str
            JSON string representation
        """
        import json
        data = self.to_json_dict(include_arrays=include_arrays, include_metadata=include_metadata)
        return json.dumps(data, indent=indent)
    
    def to_summary_dict(self):
        """
        Convert ResponseDataNode to a summary dictionary (no full arrays).
        
        Returns
        -------
        dict
            Summary dictionary with key simulation metrics
        """
        sim_info = self.get_simulation_info()
        result = {
            "status": "success",
            "simulation_summary": {
                "n_steps": sim_info.get('n_steps', len(self.t_t)),
                "time_range": sim_info.get('time_range', [float(self.t_t[0]), float(self.t_t[-1])]),
                "eps_range": sim_info.get('eps_range', [float(np.min(self.eps_t)), float(np.max(self.eps_t))]),
                "sig_range": sim_info.get('sig_range', [float(np.min(self.sig_t)), float(np.max(self.sig_t))]),
                "max_strain": float(np.max(self.eps_t)),
                "max_stress": float(np.max(self.sig_t)),
                "final_strain": float(self.eps_t[-1, 0]) if self.eps_t.ndim > 1 else float(self.eps_t[-1]),
                "final_stress": float(self.sig_t[-1, 0, 0]) if self.sig_t.ndim > 2 else float(self.sig_t[-1])
            },
            "available_variables": {
                "internal_variables": list(self.Eps_t.keys()),
                "thermodynamic_forces": list(self.Sig_t.keys())
            },
            "aiida_info": {
                "uuid": str(self.uuid),
                "is_stored": self.is_stored,
                "creation_time": self.ctime.isoformat() if hasattr(self, 'ctime') and self.ctime else None
            }
        }
        
        # Add final values of internal variables
        result["final_internal_variables"] = {}
        for var_name, var_data in self.Eps_t.items():
            if isinstance(var_data, np.ndarray):
                if var_data.ndim == 1:
                    result["final_internal_variables"][var_name] = float(var_data[-1])
                elif var_data.ndim >= 2:
                    final_val = var_data[-1].flatten()
                    if len(final_val) == 1:
                        result["final_internal_variables"][var_name] = float(final_val[0])
                    else:
                        result["final_internal_variables"][var_name] = final_val.tolist()
        
        return result
    
    def format_for_output(self, output_format='json', detailed=False):
        """
        Format ResponseDataNode for different output types.
        
        Parameters
        ----------
        output_format : str
            Output format: 'json', 'json_detailed', or 'node'
        detailed : bool
            Include full arrays (for JSON formats)
            
        Returns
        -------
        dict or ResponseDataNode
            Formatted data ready for output or transfer
        """
        if output_format == 'node':
            return self  # Return self for direct AiiDA usage
        elif output_format == 'json_detailed' or (output_format == 'json' and detailed):
            return self.to_json_dict(include_arrays=True, include_metadata=True)
        elif output_format == 'json':
            return self.to_summary_dict()
        else:
            raise ValueError(f"Unsupported output format: {output_format}. "
                           f"Supported: 'json', 'json_detailed', 'node'")

# Factory function for creating nodes
def create_response_data_node(response_data: 'ResponseData', 
                             simulation_metadata: Optional[Dict[str, Any]] = None,
                             store: bool = True,
                             auto_load_profile: bool = True) -> ResponseDataNode:
    """
    Factory function to create and optionally store a ResponseDataNode.
    
    Parameters
    ----------
    response_data : ResponseData
        The active simulation data
    simulation_metadata : dict, optional
        Additional simulation metadata
    store : bool, default True
        Whether to store the node in the database immediately
    auto_load_profile : bool, default True
        Whether to automatically load AiiDA profile if needed
        
    Returns
    -------
    ResponseDataNode
        Created (and optionally stored) node
        
    Raises
    ------
    ImportError
        If AiiDA is not available
    RuntimeError
        If AiiDA profile cannot be loaded
    """
    if not AIIDA_AVAILABLE:
        raise ImportError("AiiDA not available. Install aiida-core to create ResponseDataNode.")
    
    # Handle profile loading
    if not AIIDA_PROFILE_READY and auto_load_profile:
        try:
            aiida.load_profile()
        except Exception as e:
            raise RuntimeError(
                f"AiiDA profile not loaded and auto-load failed: {e}\n"
                "Please load a profile with aiida.load_profile() or set auto_load_profile=False."
            )
    elif not AIIDA_PROFILE_READY:
        raise RuntimeError(
            "AiiDA profile not loaded. Please load a profile with aiida.load_profile() "
            "or set auto_load_profile=True."
        )
        
    node = ResponseDataNode.from_response_data(response_data, simulation_metadata)
    
    if store:
        node.store()
    
    return node



import numpy as np
import numpy.typing as npt
import traits.api as tr
import bmcs_utils.api as bu
from typing import Dict, Any, List, Union, Optional, TYPE_CHECKING
from .response_data_viz import ResponseDataVisualizationMixin

if TYPE_CHECKING:
    from .gsm_engine import GSMEngine

class ResponseDataContainer:
    def __init__(self, data_dict: Dict[str, Any]) -> None:
        self._data = data_dict
        
    def __getitem__(self, key: str) -> Any:
        return self._data[key]
        
    def __getattr__(self, name: str) -> Any:
        if name in self._data:
            return self._data[name]
        raise AttributeError(f"No attribute named {name}")
        
    def keys(self) -> Any:
        return self._data.keys()
        
    def items(self) -> Any:
        return self._data.items()
        
    def values(self) -> Any:
        return self._data.values()

    def __len__(self) -> int:
        return len(self._data)

class ResponseData(bu.Model, ResponseDataVisualizationMixin):
    """
    Flexible container for simulation response data and metadata from GSMModel.

    This class encapsulates the time history of simulation results, including
    time, control variables, conjugate variables, internal variables, and their
    metadata (names, shapes, etc.). It provides methods to extract, reshape,
    and access the data in a form suitable for postprocessing, plotting, or storage.
    """

    t_t = tr.Array()
    eps_t = tr.Array()
    sig_t = tr.Array()
    Eps_t_flat = tr.Array()
    Sig_t_flat = tr.Array()
    Eps_vars = tr.Tuple()
    Sig_vars = tr.Tuple()
    Eps_codenames = tr.Tuple()
    Sig_codenames = tr.Tuple()
    Eps_t = tr.Any()
    Sig_t = tr.Any()
    iter_t = tr.Array()
    lam_t = tr.Array()

    @classmethod
    def from_engine_response(cls, engine: 'GSMEngine', response: tuple) -> 'ResponseData':
        """
        Construct a ResponseData object from the output tuple of GSMEngine.get_response.

        Parameters
        ----------
        engine : GSMEngine
            The GSMEngine instance (F_engine or G_engine) used for the simulation.
        response : tuple
            The tuple returned by engine.get_response.

        Returns
        -------
        ResponseData
            An instance populated with data and metadata.
        """
        t_t, eps_t, sig_t, Eps_t, Sig_t, iter_t, lam_t, _ = response
        # Flatten Eps_t and Sig_t to (n_steps, n_flat)
        Eps_t_flat_local = Eps_t.reshape(Eps_t.shape[0], -1)
        Sig_t_flat_local = Sig_t.reshape(Sig_t.shape[0], -1)
        # Get variable metadata from the GSMEngine
        Eps_vars = engine.Eps_vars
        Sig_vars = engine.Sig_vars
    
        Eps_codenames = [getattr(var, 'codename', str(var)) for var in Eps_vars]
        Sig_codenames = [getattr(var, 'codename', str(var)) for var in Sig_vars]
        
        # Try to get codenames from the parent GSMDef class if available
        gsm_def = getattr(engine, 'gsm_def', None)
        if gsm_def is None:
            # Try to infer from the engine's module (works if engine is a class attribute)
            import sys
            for obj in sys.modules[engine.__class__.__module__].__dict__.values():
                if hasattr(obj, 'F_engine') and (obj.F_engine is engine or getattr(obj, 'G_engine', None) is engine):
                    gsm_def = obj
                    break

        # Build dicts for variable access by codename (supporting scalars, vectors, tensors)
        Eps_t_dict: Dict[str, npt.NDArray[np.float64]] = {}
        col = 0
        for var in Eps_vars:
            var_shape = var.shape if hasattr(var, 'shape') else ()
            n_flat = int(np.prod(var_shape)) if var_shape else 1
            # Use codename attribute from var
            codename = getattr(var, 'codename', str(var))
            Eps_t_dict[codename] = Eps_t_flat_local[:, col:col + n_flat].reshape((Eps_t_flat_local.shape[0],) + tuple(var_shape))
            col += n_flat

        Sig_t_dict: Dict[str, npt.NDArray[np.float64]] = {}
        col = 0
        for var in Sig_vars:
            var_shape = var.shape if hasattr(var, 'shape') else ()
            n_flat = int(np.prod(var_shape)) if var_shape else 1
            codename = getattr(var, 'codename', str(var))
            Sig_t_dict[codename] = Sig_t_flat_local[:, col:col + n_flat].reshape((Sig_t_flat_local.shape[0],) + tuple(var_shape))
            col += n_flat

        return cls(
            t_t=t_t,
            eps_t=eps_t,
            sig_t=sig_t,
            Eps_t_flat=Eps_t_flat_local,
            Sig_t_flat=Sig_t_flat_local,
            Eps_vars=Eps_vars,
            Sig_vars=Sig_vars,
            Eps_codenames=Eps_codenames,
            Sig_codenames=Sig_codenames,
            Eps_t=ResponseDataContainer(Eps_t_dict),
            Sig_t=ResponseDataContainer(Sig_t_dict),
            iter_t=iter_t,
            lam_t=lam_t
        )

    def get_internal_var(self, key: Union[str, int]) -> npt.NDArray[np.float64]:
        """
        Extract the internal variable as a time series, accessed by codename or index.

        Parameters
        ----------
        key : str or int
            Codename of the internal variable (as defined in GSMDef), or index.

        Returns
        -------
        arr : np.ndarray
            Array of shape (n_steps,) for scalar variables.
        """
        if isinstance(key, int):
            return self.get_internal_var_by_index(key)
        return self.Eps_t[key]

    def get_internal_var_by_index(self, i: int) -> npt.NDArray[np.float64]:
        var_shape = self.Eps_vars[i].shape
        start = sum(int(np.prod(v.shape)) for v in self.Eps_vars[:i])
        end = start + int(np.prod(var_shape))
        flat = self.Eps_t_flat[:, start:end]
        return flat.reshape((flat.shape[0],) + var_shape)

    def get_conj_var(self, key: Union[str, int]) -> npt.NDArray[np.float64]:
        """
        Extract the conjugate variable (thermodynamic force) as a time series, accessed by codename or index.

        Parameters
        ----------
        key : str or int
            Codename of the conjugate variable (as defined in GSMDef), or index.

        Returns
        -------
        arr : np.ndarray
            Array of shape (n_steps,) for scalar variables.
        """
        if isinstance(key, int):
            return self.get_conj_var_by_index(key)
        return self.Sig_t[key]

    def get_conj_var_by_index(self, i: int) -> npt.NDArray[np.float64]:
        var_shape = self.Sig_vars[i].shape
        start = sum(int(np.prod(v.shape)) for v in self.Sig_vars[:i])
        end = start + int(np.prod(var_shape))
        flat = self.Sig_t_flat[:, start:end]
        return flat.reshape((flat.shape[0],) + var_shape)

    def to_dict(self) -> Dict[str, Any]:
        """
        Return all data and metadata as a dictionary for storage or further processing.
        """
        return dict(
            t_t=self.t_t,
            eps_t=self.eps_t,
            sig_t=self.sig_t,
            Eps_t_flat=self.Eps_t_flat,
            Sig_t_flat=self.Sig_t_flat,
            Eps_vars=self.Eps_vars,
            Sig_vars=self.Sig_vars,
            Eps_codenames=self.Eps_codenames,
            Sig_codenames=self.Sig_codenames,
            Eps_t=self.Eps_t,
            Sig_t=self.Sig_t,
            iter_t=self.iter_t,
            lam_t=self.lam_t
        )

    def __getitem__(self, key: str) -> Any:
        """
        Allow dict-like access to main arrays.
        """
        return getattr(self, key)

    def __repr__(self) -> str:
        return (
            f"ResponseData(t_t.shape={self.t_t.shape}, "
            f"eps_t.shape={self.eps_t.shape}, "
            f"sig_t.shape={self.sig_t.shape}, "
            f"Eps_t_flat.shape={self.Eps_t_flat.shape}, "
            f"Sig_t_flat.shape={self.Sig_t_flat.shape}, "
            f"Eps_vars={self.Eps_vars}, "
            f"Sig_vars={self.Sig_vars}, "
            f"Eps_codenames={self.Eps_codenames}, "
            f"Sig_codenames={self.Sig_codenames}, "
            f"Eps_t keys={list(self.Eps_t.keys())}, "
            f"Sig_t keys={list(self.Sig_t.keys())})"
        )
        return getattr(self, key)

    def __repr__(self):
        return (
            f"ResponseData(t_t.shape={self.t_t.shape}, "
            f"eps_t.shape={self.eps_t.shape}, "
            f"sig_t.shape={self.sig_t.shape}, "
            f"Eps_t_flat.shape={self.Eps_t_flat.shape}, "
            f"Sig_t_flat.shape={self.Sig_t_flat.shape}, "
            f"Eps_vars={self.Eps_vars}, "
            f"Sig_vars={self.Sig_vars}, "
            f"Eps_codenames={self.Eps_codenames}, "
            f"Sig_codenames={self.Sig_codenames}, "
            f"Eps_t keys={list(self.Eps_t.keys())}, "
            f"Sig_t keys={list(self.Sig_t.keys())})"
        )



#!/usr/bin/env python3
"""
Unified GSM Definition Registry

This module provides automatic discovery of GSM definitions by scanning for files
matching the pattern 'gsm1d_*.py' and importing GSMDef classes only when accessed.

GSM Definitions define the constitutive behavior and can be instantiated with 
specific material parameters to create GSM Models for particular loading scenarios.

Key features:
- Path-based discovery (safe for CLI)
- Lazy loading of GSMDef classes (import only when needed)
- Single unified approach for both CLI and notebook use
"""

import os
import importlib
import inspect
from typing import Dict, List, Type, Tuple, Optional, Union, Any


class LazyGSMDefRegistry:
    """
    Registry that discovers GSM definitions by file scanning and imports GSMDef classes on-demand.
    
    GSM Definitions are classes that define constitutive behavior. They can be instantiated
    with specific material parameters to create GSM Models for loading scenarios.
    
    This approach combines the safety of path-based discovery (no upfront imports)
    with the functionality of class-based access (import when needed).
    """
    
    def __init__(self, search_dir: Optional[str] = None, debug: bool = False):
        # Default to the models directory relative to this file
        if search_dir is None:
            current_dir = os.path.dirname(__file__)
            search_dir = os.path.join(current_dir, '..', 'models')
        self.search_dir = search_dir
        self.debug = debug
        self._def_paths = {}  # def_name -> module_path
        self._access_keys = {}  # all_keys -> def_name  
        self._loaded_def_classes = {}  # def_name -> actual_GSMDef_class (cache)
        self._discover_definitions()
    
    def _discover_definitions(self) -> None:
        """Discover GSM definitions by scanning files (no imports yet)"""
        # Get list of potential GSM definition files
        gsm_files = [f for f in os.listdir(self.search_dir) 
                     if f.endswith('.py') and f.startswith('gsm1d_')]
        
        if self.debug:
            print(f"Found {len(gsm_files)} potential GSM definition files:")
            for file_name in gsm_files:
                print(f"  {file_name}")
        
        # Extract definition information from file names
        for file_name in gsm_files:
            module_name = file_name[:-3]  # Remove .py extension
            
            # Infer definition name from module name: gsm1d_ed.py -> GSM1D_ED
            if module_name.startswith('gsm1d_'):
                mechanism_part = module_name[6:]  # Remove 'gsm1d_'
                def_name = f"GSM1D_{mechanism_part.upper()}"
                module_path = f"bmcs_matmod.gsm_lagrange.models.{module_name}"
                
                # Store the definition path
                self._def_paths[def_name] = module_path
                
                # Create all access variants
                self._access_keys[def_name] = def_name
                self._access_keys[def_name.lower()] = def_name
                self._access_keys[mechanism_part.upper()] = def_name
                self._access_keys[mechanism_part.lower()] = def_name
                
                if self.debug:
                    print(f"✓ Detected definition: {def_name}")
                    print(f"  Module: {module_name}")
                    print(f"  Mechanism: {mechanism_part.upper()}")
        
        if self.debug:
            print(f"Total discovered definitions: {len(self._def_paths)}")
            print(f"Total access keys: {len(self._access_keys)}")
    
    def _load_def_class(self, def_name: str) -> Type:
        """Lazily load the actual GSMDef class when needed"""
        if def_name in self._loaded_def_classes:
            return self._loaded_def_classes[def_name]
        
        if def_name not in self._def_paths:
            raise ValueError(f"GSM definition '{def_name}' not found")
        
        module_path = self._def_paths[def_name]
        
        try:
            # Import GSMDef for type checking
            try:
                from .gsm_def import GSMDef
            except ImportError:
                try:
                    from gsm_def import GSMDef
                except ImportError:
                    # Try absolute import
                    from bmcs_matmod.gsm_lagrange.core.gsm_def import GSMDef
            
            # Import the module
            module = importlib.import_module(module_path)
            
            # Find the GSMDef subclass in the module
            for name, obj in inspect.getmembers(module, inspect.isclass):
                if (issubclass(obj, GSMDef) and 
                    obj is not GSMDef and 
                    name == def_name):
                    self._loaded_def_classes[def_name] = obj
                    if self.debug:
                        print(f"✓ Loaded GSMDef class: {def_name} from {module_path}")
                    return obj
            
            raise ValueError(f"No GSMDef subclass '{def_name}' found in {module_path}")
            
        except Exception as e:
            raise ImportError(f"Failed to load GSM definition {def_name} from {module_path}: {e}")
    
    def get_def_names(self) -> List[str]:
        """Get list of discovered GSM definition names (no imports)"""
        return sorted(self._def_paths.keys())
    
    def get_access_keys(self) -> List[str]:
        """Get all available access keys including aliases (no imports)"""
        return sorted(self._access_keys.keys())
    
    def has_def(self, key: str) -> bool:
        """Check if a GSM definition exists (no imports)"""
        return key in self._access_keys
    
    def get_module_path(self, key: str) -> Optional[str]:
        """Get module path for a GSM definition (no imports)"""
        def_name = self._access_keys.get(key)
        if def_name:
            return self._def_paths[def_name]
        return None
    
    def get_def_class(self, key: str) -> Type:
        """Get the actual GSMDef class (imports on first access)"""
        def_name = self._access_keys.get(key)
        if not def_name:
            available = list(self._access_keys.keys())
            raise ValueError(f"GSM definition '{key}' not found. Available: {available}")
        
        return self._load_def_class(def_name)
    
    def get_all_def_classes(self) -> Dict[str, Type]:
        """Get all GSMDef classes as a dictionary (imports all)"""
        result = {}
        for key, def_name in self._access_keys.items():
            if def_name not in result:  # Avoid duplicates
                result[key] = self._load_def_class(def_name)
        return result
    
    def list_defs_with_classes(self) -> List[Tuple[str, Type]]:
        """Get list of (def_name, GSMDef_class) tuples (imports all)"""
        result = []
        for def_name in self.get_def_names():
            def_class = self._load_def_class(def_name)
            result.append((def_name, def_class))
        return result


# Global registry instance
_registry = None


def _get_registry(debug: bool = False) -> LazyGSMDefRegistry:
    """Get or create the global registry instance"""
    global _registry
    if _registry is None:
        _registry = LazyGSMDefRegistry(debug=debug)
    return _registry


# Public API functions that match the original interfaces

def discover_gsm_defs(search_dir: Optional[str] = None, debug: bool = False) -> Tuple[List[Tuple[str, Type]], Dict[str, Type]]:
    """
    Discover GSM definitions (compatible with original function).
    
    Returns:
    --------
    Tuple containing:
    - List of (def_name, GSMDef_class) tuples
    - Registry dictionary with multiple access keys
    """
    registry = LazyGSMDefRegistry(search_dir, debug)
    defs = registry.list_defs_with_classes()
    classes_dict = registry.get_all_def_classes()
    return defs, classes_dict


def get_gsm_defs(debug: bool = False) -> Dict[str, Type]:
    """
    Get GSM defs registry dictionary (compatible with original function).
    
    Returns a dictionary that allows access by:
    - Full class name (e.g., 'GSM1D_ED')
    - Lowercase class name (e.g., 'gsm1d_ed')
    - Mechanism name (e.g., 'ED')
    - Lowercase mechanism name (e.g., 'ed')
    """
    registry = _get_registry(debug)
    return registry.get_all_def_classes()


def list_gsm_defs(debug: bool = False) -> List[Tuple[str, Type]]:
    """
    Get list of discovered GSM defs (compatible with original function).
    
    Returns:
    --------
    List of (def_name, GSMDef_class) tuples sorted by name
    """
    registry = _get_registry(debug)
    return registry.list_defs_with_classes()


# CLI-friendly functions (no imports until needed)

def get_available_gsm_defs(debug: bool = False) -> Tuple[List[str], Dict[str, str]]:
    """Get available GSM definitions without importing them (for CLI)"""
    registry = _get_registry(debug)
    defs = registry.get_def_names()
    paths = {key: registry.get_module_path(key) for key in registry.get_access_keys()}
    return defs, paths


def check_gsm_def_exists(def_key: str, debug: bool = False) -> bool:
    """Check if a GSM definition exists without importing it (for CLI)"""
    registry = _get_registry(debug)
    return registry.has_def(def_key)


def get_gsm_def_module_path(def_key: str, debug: bool = False) -> Optional[str]:
    """Get the module path for a GSM definition (for CLI)"""
    registry = _get_registry(debug)
    return registry.get_module_path(def_key)


def get_gsm_def_class(def_key: str, debug: bool = False) -> Type:
    """Get the actual GSMDef class (imports on first access)"""
    registry = _get_registry(debug)
    return registry.get_def_class(def_key)


# Display and utility functions

def print_gsm_defs(debug: bool = True) -> None:
    """
    Print discovered GSM definitions with details.
    """
    registry = _get_registry(debug)
    defs = registry.list_defs_with_classes()
    
    if not debug:
        print("\nDiscovered GSM Definitions:")
        print("=" * 60)
        for i, (name, obj) in enumerate(defs, 1):
            print(f"{i:2d}. {name}")
            
            # Extract mechanism from name
            if name.startswith('GSM1D_'):
                mechanism = name[6:]  # Remove 'GSM1D_'
                print(f"    Mechanism: {mechanism}")
            
            print(f"    GSMDef Class: {obj}")
            print()


# Mechanism descriptions for display purposes
MECHANISM_DESCRIPTIONS = {
    'ED': 'Elasto-Damage Model',
    'VE': 'Visco-Elastic Model',
    'VED': 'Visco-Elasto-Damage Model', 
    'EP': 'Elasto-Plastic Model',
    'EPD': 'Elasto-Plastic-Damage Model',
    'EVP': 'Elasto-Visco-Plastic Model',
    'EVPD': 'Elasto-Visco-Plastic-Damage Model',
    'VEVP': 'Visco-Elasto-Visco-Plastic Model',
    'VEVPD': 'Visco-Elasto-Visco-Plastic-Damage Model'
}


def get_mechanism_description(mechanism: str) -> str:
    """Get description for a mechanism code."""
    return MECHANISM_DESCRIPTIONS.get(mechanism.upper(), f"{mechanism} Model")





if __name__ == "__main__":
    # When run as script, demonstrate both modes
    print("Unified GSM Definition Registry Demo")
    print("=" * 50)
    
    print("\n1. Path-based discovery (no imports):")
    registry = LazyGSMDefRegistry(debug=True)
    print(f"Found definitions: {registry.get_def_names()}")
    print(f"Access keys: {registry.get_access_keys()}")
    
    print("\n2. On-demand GSMDef class loading:")
    if registry.get_def_names():
        first_def = registry.get_def_names()[0]
        print(f"Loading GSMDef class for: {first_def}")
        try:
            def_class = registry.get_def_class(first_def)
            print(f"✓ Successfully loaded: {def_class}")
        except Exception as e:
            print(f"✗ Failed to load: {e}")
    
    print("\n3. Traditional API compatibility:")
    try:
        defs, registry_dict = discover_gsm_defs(debug=False)
        print(f"✓ discover_gsm_defs: {len(defs)} definitions")
        
        defs_dict = get_gsm_defs(debug=False)
        print(f"✓ get_gsm_defs: {len(defs_dict)} entries")
        
        defs_list = list_gsm_defs(debug=False)
        print(f"✓ list_gsm_defs: {len(defs_list)} definitions")
        
    except Exception as e:
        print(f"✗ Traditional API failed: {e}")
